import sys

from PyQt5.QtWidgets import QStyleOptionViewItem

import pandasgui
import os

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt

from pandasgui.store import PandasGuiDataFrameStore
from pandasgui.widgets.collapsible_panel import CollapsiblePanel

from pandasgui.widgets.func_ui import FuncUi, Schema

import logging

from pandasgui.widgets.grapher import TypePicker

logger = logging.getLogger(__name__)


class Reshaper(QtWidgets.QWidget):
    def __init__(self, pgdf: PandasGuiDataFrameStore):
        super().__init__()

        self.pgdf = PandasGuiDataFrameStore.cast(pgdf)

        self.setWindowTitle("Graph Builder")

        # Dropdown to select plot type
        self.type_picker = TypePicker()

        self.error_console = QtWidgets.QTextEdit()
        self.error_console.setReadOnly(True)
        self.error_console_wrapper = CollapsiblePanel("Reshaper Console")
        self.error_console_wrapper.setContent(self.error_console)
        self.error_console_wrapper.setChecked(False)

        for schema in schemas:
            icon = QtGui.QIcon(schema.icon_path)
            text = schema.label
            item = QtWidgets.QListWidgetItem(icon, text)
            self.type_picker.addItem(item)

        self.func_ui = FuncUi(pgdf=pgdf, schema=Schema())

        # Layouts
        self.plot_splitter = QtWidgets.QSplitter(Qt.Horizontal)

        self.plot_splitter.addWidget(self.func_ui)

        self.plot_splitter.setStretchFactor(1, 1)

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.type_picker)
        self.layout.addWidget(self.plot_splitter)
        self.layout.addWidget(self.error_console_wrapper)
        self.setLayout(self.layout)

        # Size policies
        self.func_ui.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.plot_splitter.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.error_console.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Minimum)

        # Initial selection
        self.type_picker.setCurrentRow(0)
        self.on_type_changed()

        # Signals
        self.type_picker.itemSelectionChanged.connect(self.on_type_changed)
        self.func_ui.finished.connect(self.on_dragger_finished)
        self.func_ui.saving.connect(self.on_dragger_saving)

    def on_type_changed(self):
        if len(self.type_picker.selectedItems()) == 0:
            return

        self.selected_plot_label = self.type_picker.selectedItems()[0].text()
        self.current_schema = next(filter(lambda schema: schema.label == self.selected_plot_label, schemas))
        arg_list = [arg.arg_name for arg in self.current_schema.args]

        self.func_ui.set_schema(self.current_schema)

    def on_dragger_saving(self):
        options = QtWidgets.QFileDialog.Options()
        # using native widgets so it matches the PNG download button

        filename, _ = QtWidgets.QFileDialog().getSaveFileName(self, "Save plot to", "", "HTML Files (*.html)",
                                                              options=options)
        if filename:
            if filename[-5:] != ".html":
                filename += ".html"
            self.fig.write_html(filename)
            self.pgdf.add_history_item("Reshaper",
                                       f"fig.write_html('{filename})'")

    def on_dragger_finished(self):
        kwargs = self.func_ui.get_data()
        func = self.current_schema.function

        try:
            self.result = func(data_frame=self.pgdf.df, **kwargs)
            self.pgdf.store.add_dataframe(self.result)

            self.error_console.setText("")
            self.error_console_wrapper.setTitle("Reshaper Console")
        except Exception as e:
            import traceback
            self.error_console.setText(traceback.format_exc())
            self.error_console_wrapper.setTitle("*** Reshaper Console ***")

    def set_state(self, plot_type, dct):
        self.type_picker.setCurrentIndex(self.type_picker.model().index(
            [schema.name for schema in schemas].index(plot_type), 0
        ))
        self.on_type_changed()
        print(self.current_schema)
        self.func_ui.set_data(dct)


def clear_layout(layout):
    for i in reversed(range(layout.count())):
        layout.itemAt(i).widget().setParent(None)


from pandasgui import jotly

schemas = [Schema(name="pivot",
                  label="Pivot",
                  function=jotly.pivot,
                  icon_path=os.path.join(pandasgui.__path__[0], "resources/images/draggers/pivot.svg"),
                  ),
           Schema(name="melt",
                  label="Melt",
                  function=jotly.melt,
                  icon_path=os.path.join(pandasgui.__path__[0], "resources/images/draggers/unpivot.svg"),
                  ),
           Schema(name="merge",
                  label="Merge",
                  function=jotly.merge,
                  icon_path=os.path.join(pandasgui.__path__[0], "resources/images/draggers/join.svg"),
                  ),
           Schema(name="concat",
                  label="Concat",
                  function=jotly.concat,
                  icon_path=os.path.join(pandasgui.__path__[0], "resources/images/draggers/concat.svg"),
                  ),
           ]

if __name__ == "__main__":
    from pandasgui.datasets import pokemon

    app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv)

    pgdf = PandasGuiDataFrameStore(pokemon)
    gb = Reshaper(pgdf)
    gb.show()

    # gb.set_state('scatter', {'y': 'Attack', 'x': 'Defense'})

    app.exec_()



import re
from PyQt5 import QtWidgets


class SearchableListWidget(QtWidgets.QWidget):
    def __init__(self, items, parent=None):
        super().__init__(parent)

        self.search_bar = QtWidgets.QLineEdit()
        self.search_bar.textChanged.connect(self.filter)

        self.list_widget = QtWidgets.QListWidget()

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self.search_bar)
        layout.addWidget(self.list_widget)
        self.setLayout(layout)

        self.initial_items = items
        self.set_items(items)

    def filter(self):
        filtered_items = [
            item
            for item in self.initial_items
            if re.search(self.search_bar.text().lower(), item.lower())
        ]

        self.set_items(filtered_items)

    def get_items(self):
        return [
            str(self.list_widget.item(i).text())
            for i in range(self.list_widget.count())
        ]

    def set_items(self, items):
        self.list_widget.clear()
        for name in items:
            self.list_widget.addItem(name)


if __name__ == "__main__":
    import sys
    from PyQt5.QtWidgets import QApplication

    app = QApplication(sys.argv)

    test = SearchableListWidget(['cat', 'dog', 'hotdog', 'catch'])
    test.show()

    sys.exit(app.exec_())



import ast
import pprint

import typing
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt
import sys
from pandasgui.store import SETTINGS_STORE, SettingsStore
from pandasgui.widgets import base_widgets
from typing_extensions import Literal, get_origin


class SettingsEditor(QtWidgets.QWidget):
    def __init__(self, settings: SettingsStore, parent=None):
        super().__init__(parent)

        self.settings = settings
        self.tree = base_widgets.QTreeWidget()
        self.tree.setHeaderLabels(['Name', 'Value'])
        self.tree.setRootIsDecorated(False)

        finish_button = QtWidgets.QPushButton('Finish')
        finish_button.clicked.connect(self.finish)

        reset_button = QtWidgets.QPushButton('Reset To Defaults')
        reset_button.clicked.connect(self.reset_to_defaults)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.tree)
        layout.addWidget(finish_button)
        layout.addWidget(reset_button)

        self.init_tree()

        self.setLayout(layout)

    def init_tree(self):

        self.tree.clear()

        # Dont display these settings because they have effects during the GUI initialization so changing them doesn't work
        settings_to_show = {k: v for k, v in self.settings.__dict__.items() if k not in ['block']}

        # Make a QTreeWidget with widgets in col2 that set its value
        for ix, (setting_name, setting) in enumerate(settings_to_show.items()):

            item = QtWidgets.QTreeWidgetItem(self.tree, [setting_name])
            item.setData(1, Qt.UserRole, setting.value)
            if setting.dtype == str:

                def setter(new_val, item=item):
                    try:
                        item.setData(1, Qt.UserRole, new_val)
                    except:
                        pass

                widget = QtWidgets.QLineEdit(setting.value)
                widget.textChanged.connect(setter)

            elif setting.dtype == dict:

                def setter(new_val, item=item):
                    try:
                        item.setData(1, Qt.UserRole, dict(ast.literal_eval(new_val)))
                    except:
                        pass

                widget = QtWidgets.QPlainTextEdit(pprint.pformat(setting.value))
                widget.textChanged.connect(lambda widget=widget: setter(widget.toPlainText()))

            elif setting.dtype == bool:

                def setter(new_val, item=item):
                    try:
                        item.setData(1, Qt.UserRole, new_val == Qt.Checked)
                    except:
                        pass

                widget = QtWidgets.QCheckBox(setting.label)
                widget.setCheckState(Qt.Checked if setting.value else Qt.Unchecked)
                widget.stateChanged.connect(setter)

            elif get_origin(setting.dtype) == Literal:

                def setter(new_val, item=item):
                    try:
                        item.setData(1, Qt.UserRole, new_val)
                    except:
                        pass

                widget = QtWidgets.QComboBox()
                widget.addItems(typing.get_args(setting.dtype))
                widget.setCurrentText(setting.value)
                widget.currentTextChanged.connect(setter)

            else:
                raise TypeError

            item.treeWidget().setItemWidget(item, 1, widget)

    def finish(self):
        # Get args from destination UI
        root = self.tree.invisibleRootItem()
        for i in range(root.childCount()):
            item = root.child(i)
            setting_name = item.text(0)
            setting_value = item.data(1, Qt.UserRole)
            self.settings[setting_name].value = setting_value
        self.settings.settingsChanged.emit()

    def reset_to_defaults(self):
        self.settings.reset_to_defaults()
        self.init_tree()
        self.finish()


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)


    d = SettingsEditor(SETTINGS_STORE)
    d.show()

    sys.exit(app.exec_())



import os
from appdirs import user_data_dir

LOCAL_DATA_DIR = os.path.join(user_data_dir(), "pandasgui")
LOCAL_DATASET_DIR = os.path.join(LOCAL_DATA_DIR, "dataset_files")

os.makedirs(LOCAL_DATA_DIR, exist_ok=True)
os.makedirs(LOCAL_DATASET_DIR, exist_ok=True)



import os
import sys

from PyQt5 import QtCore, QtGui, QtWidgets

from pandasgui.store import PandasGuiDataFrameStore, SETTINGS_STORE
import pandasgui

import logging

from pandasgui.utility import resize_widget
from pandasgui.widgets.python_highlighter import PythonHighlighter

logger = logging.getLogger(__name__)


class CodeHistoryViewer(QtWidgets.QFrame):
    def __init__(self, pgdf: PandasGuiDataFrameStore):
        super().__init__()
        pgdf = PandasGuiDataFrameStore.cast(pgdf)
        pgdf.filter_viewer = self
        self.pgdf = pgdf

        resize_widget(self, 0.5, 0.5)

        self.textbox = QtWidgets.QPlainTextEdit()
        self.textbox.setReadOnly(True)
        self.textbox.setLineWrapMode(self.textbox.NoWrap)
        self.highlighter = PythonHighlighter(self.textbox.document(),
                                             dark=SETTINGS_STORE.theme.value == 'dark')

        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.textbox)
        self.setLayout(self.layout)

        self.setWindowTitle(f"Code Export ({self.pgdf.name})")

        self.refresh()

    def refresh(self):
        code_history = self.pgdf.code_export()
        self.textbox.setPlainText(code_history)

    def closeEvent(self, a0: QtGui.QCloseEvent) -> None:
        self.hide()


if __name__ == "__main__":
    # Create a QtWidgets.QApplication instance or use the existing one if it exists
    app = QtWidgets.QApplication(sys.argv)
    from pandasgui.datasets import pokemon

    pokemon = PandasGuiDataFrameStore.cast(pokemon)
    pokemon.parse_all_dates()
    pokemon.sort_column(0)
    fv = CodeHistoryViewer(pokemon)
    fv.show()
    app.exec_()



import sys

from PyQt5 import QtWidgets



import logging
logger = logging.getLogger(__name__)

class InputDialog(QtWidgets.QDialog):
    """
    Example usage:
    dialog = InputDialog("Add Column", {"Column Title": str, "Column Formula": str})
    column_title, column_formula = dialog
    """
    def __init__(self, title, schema, parent=None):
        super().__init__(parent)

        layout = QtWidgets.QFormLayout(self)

        self.fields = []
        for field_name, field_type in schema.items():
            if field_type == str:
                widget = QtWidgets.QLineEdit(self)
            if field_type == bool:
                pass  # TODO
            self.fields.append(widget)
            layout.addRow(field_name, self.first)


        buttonBox = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel, self);

        layout.addRow("Second text", self.second)
        layout.addWidget(buttonBox)

        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

    def getInputs(self):
        return (self.first.text(), self.second.text())

if __name__ == '__main__':

    import sys
    app = QtWidgets.QApplication(sys.argv)
    dialog = InputDialog()
    if dialog.exec():
        print(dialog.getInputs())
    exit(0)



import sys
from typing import Union, List
from PyQt5 import QtCore, QtGui, QtWidgets

from pandasgui.store import PandasGuiStoreItem
from pandasgui.utility import summarize_json, traverse_tree_widget
from pandasgui.widgets import base_widgets

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

class JsonViewer(QtWidgets.QWidget, PandasGuiStoreItem):

    def __init__(self, jdata: Union[list, dict], parent=None):
        super().__init__(parent)

        self.jdata = jdata

        self.search_mode: Literal['filter', 'highlight'] = 'filter'

        self.search_mode_picker = QtWidgets.QComboBox()
        self.search_mode_picker.addItems(['filter', 'highlight'])
        self.search_mode_picker.currentTextChanged.connect(lambda text: [self.__setattr__('search_mode', text),
                                                                         self.find()])

        self.find_box = QtWidgets.QLineEdit()
        self.find_box.returnPressed.connect(self.find)
        self.find_box.textChanged.connect(self.find)
        self.find_box.setPlaceholderText("Find regex")

        self.tree_widget = base_widgets.QTreeWidget()
        self.tree_widget.setHeaderLabels(["Key", "Value"])

        root_item = self.tree_widget.invisibleRootItem()
        self.recurse_jdata(jdata, root_item)
        self.tree_widget.addTopLevelItem(root_item)
        self.tree_widget.expandAll()

        # We call .find after collapse because it will re-expand highlighted items (matching the find_box text)
        self.expand_all = QtWidgets.QPushButton("Expand All")
        self.expand_all.clicked.connect(lambda: [self.tree_widget.expandAll(), self.find()])
        self.collapse_all = QtWidgets.QPushButton("Collapse All")
        self.collapse_all.clicked.connect(lambda: [self.tree_widget.collapseAll(), self.find()])

        top_section = QtWidgets.QHBoxLayout()
        top_section.addWidget(self.find_box)
        top_section.addWidget(self.search_mode_picker)
        top_section.addWidget(self.expand_all)
        top_section.addWidget(self.collapse_all)

        main_view_layout = QtWidgets.QVBoxLayout()
        main_view_layout.addLayout(top_section)
        main_view_layout.addWidget(self.tree_widget)

        main_view = QtWidgets.QWidget()
        main_view.setLayout(main_view_layout)

        summary_view = QtWidgets.QTextEdit()
        summary_view.setReadOnly(True)
        summary_view.setText(summarize_json(jdata))

        font = QtGui.QFont("Monospace")
        font.setStyleHint(QtGui.QFont.TypeWriter)
        summary_view.setFont(font)

        self.tabs = QtWidgets.QTabWidget()
        self.tabs.addTab(main_view, "Viewer")
        self.tabs.addTab(summary_view, "Structure")

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.tabs)

        self.setLayout(layout)
        self.resize(QtCore.QSize(400, 500))

        self.tree_widget.setSelectionMode(QtWidgets.QTreeWidget.ExtendedSelection)

    def find(self):

        text = self.find_box.text()

        if text == "":
            # Unhide and unselect all items
            for item in traverse_tree_widget(self.tree_widget):
                item.setHidden(False)
            self.tree_widget.clearSelection()
            return

        result: List[QtWidgets.QTreeWidgetItem] = []
        for col in range(self.tree_widget.columnCount()):
            result += self.tree_widget.findItems(f".*{text}.*",
                                                 QtCore.Qt.MatchRegExp | QtCore.Qt.MatchRecursive,
                                                 col)

        if self.search_mode == 'highlight':
            # Unhide all items
            for item in traverse_tree_widget(self.tree_widget):
                item.setHidden(False)

            # Highlight matching items
            self.tree_widget.clearSelection()
            self.tree_widget.setSelectionMode(QtWidgets.QTreeWidget.MultiSelection)
            for item in result:
                item.setSelected(True)
            self.tree_widget.setSelectionMode(QtWidgets.QTreeWidget.ExtendedSelection)

        elif self.search_mode == 'filter':
            # Unselect all items
            self.tree_widget.clearSelection()
            # Hide all items
            for item in traverse_tree_widget(self.tree_widget):
                item.setHidden(True)
            # Show matching items and parents
            for item in result:
                while item:
                    item.setHidden(False)
                    item.setExpanded(True)
                    item = item.parent()

        # Expand found items
        for item in result:
            while item:
                item.setExpanded(True)
                item = item.parent()

    def recurse_jdata(self, jdata, tree_widget):

        if isinstance(jdata, dict):
            items = jdata.items()
        elif isinstance(jdata, list):
            items = [(str(i), val) for i, val in enumerate(jdata)]
        else:
            raise ValueError(f"Expected dict or list, instead got {type(jdata)}")

        for key, val in items:
            text_list = []

            if isinstance(val, dict) or isinstance(val, list):
                text_list.append(key)
                row_item = QtWidgets.QTreeWidgetItem([key])
                self.recurse_jdata(val, row_item)
            else:
                text_list.append(key)
                text_list.append(str(val))
                row_item = QtWidgets.QTreeWidgetItem([key, str(val)])

            tree_widget.addChild(row_item)

    def pg_widget(self):
        return self


if "__main__" == __name__:
    app = QtWidgets.QApplication([])

    data = {
        "Employees": [
            {
                "userId": "rirani",
                "jobTitleName": "Developer",
                "firstName": "Romin",
                "lastName": "Irani",
                "preferredFullName": "Romin Irani",
                "employeeCode": "E1",
                "region": "CA",
                "phoneNumber": "408-1234567",
                "emailAddress": "romin.k.irani@gmail.com"
            },
            {
                "userId": "nirani",
                "jobTitleName": "Developer",
                "firstName": "Neil",
                "lastName": "Irani",
                "preferredFullName": "Neil Irani",
                "employeeCode": "E2",
                "region": "CA",
                "phoneNumber": "408-1111111",
                "emailAddress": "neilrirani@gmail.com"
            },
            {
                "userId": "thanks",
                "jobTitleName": "Program Directory",
                "firstName": "Tom",
                "lastName": "Hanks",
                "preferredFullName": "Tom Hanks",
                "employeeCode": "E3",
                "region": "CA",
                "phoneNumber": "408-2222222",
                "emailAddress": "tomhanks@gmail.com"
            }
        ]
    }

    example = JsonViewer(data)
    example.show()


    refs = []
    import requests

    for url in [
        # 'https://jsonplaceholder.typicode.com/posts',  # 100 posts
        # 'https://jsonplaceholder.typicode.com/comments',  # 500 comments
        # 'https://jsonplaceholder.typicode.com/albums',  # 100 albums
        # 'https://jsonplaceholder.typicode.com/photos',  # 5000 photos
        # 'https://jsonplaceholder.typicode.com/todos',  # 200 todos
        # 'https://jsonplaceholder.typicode.com/users',  # 10 users
    ]:
        data = requests.get(url).json()
        x = JsonViewer(data)
        x.show()
        refs.append(x)

    sys.exit(app.exec_())



import sys
from typing import List
from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import Qt
from pandasgui.widgets.code_history_viewer import CodeHistoryViewer

from pandasgui.widgets.containers import Container
from pandasgui.widgets.dataframe_viewer import DataFrameViewer
from pandasgui.widgets.column_viewer import ColumnViewer
from pandasgui.widgets.grapher import Grapher
from pandasgui.widgets.reshaper import Reshaper
from pandasgui.widgets.filter_viewer import FilterViewer
from pandasgui.widgets.stats_viewer import StatisticsViewer
from pandasgui.widgets.dock_widget import DockWidget
from pandasgui.store import PandasGuiDataFrameStore

import logging

logger = logging.getLogger(__name__)


class DataFrameExplorer(QtWidgets.QWidget):
    def __init__(self, pgdf: PandasGuiDataFrameStore):
        super().__init__()

        pgdf = PandasGuiDataFrameStore.cast(pgdf)
        pgdf.dataframe_explorer = self
        self.pgdf = pgdf

        ##################
        # Set up main views in Dock tabs
        self.main_window = QtWidgets.QMainWindow()
        self.docks: List[DockWidget] = []
        self.main_window.setDockOptions(
            self.main_window.GroupedDragging | self.main_window.AllowTabbedDocks | self.main_window.AllowNestedDocks)
        self.main_window.setTabPosition(Qt.AllDockWidgetAreas, QtWidgets.QTabWidget.North)

        self.dataframe_viewer = DataFrameViewer(pgdf)
        self.statistics_viewer = StatisticsViewer(pgdf)
        self.grapher = Grapher(pgdf)
        self.reshaper = Reshaper(pgdf)
        self.code_history_viewer = CodeHistoryViewer(pgdf)

        self.dataframe_dock = self.add_view(self.dataframe_viewer, "DataFrame")
        self.statistics_dock = self.add_view(self.statistics_viewer, "Statistics")
        self.grapher_dock = self.add_view(self.grapher, "Grapher")
        self.reshaper_dock = self.add_view(self.reshaper, "Reshaper")

        def set_active_tab(name):
            self.active_tab = name

        self.dataframe_dock.activated.connect(lambda: set_active_tab("DataFrame"))
        self.statistics_dock.activated.connect(lambda: set_active_tab("Statistics"))
        self.grapher_dock.activated.connect(lambda: set_active_tab("Grapher"))
        self.reshaper_dock.activated.connect(lambda: set_active_tab("Reshaper"))

        self.dataframe_viewer.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        ##################
        # Non-Dock widgets

        self.filter_viewer = FilterViewer(pgdf)
        self.column_viewer = ColumnArranger(self.pgdf)

        ##################
        # Set up overall layout

        self.splitter = QtWidgets.QSplitter(Qt.Horizontal)
        self.side_bar = QtWidgets.QSplitter(Qt.Vertical)

        self.layout = QtWidgets.QHBoxLayout()
        self.setLayout(self.layout)
        self.layout.addWidget(self.splitter)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)

        self.filter_viewer_container = Container(self.filter_viewer, "Filters")

        self.splitter.addWidget(self.side_bar)
        self.splitter.addWidget(self.main_window)
        self.side_bar.addWidget(self.filter_viewer_container)
        self.side_bar.addWidget(Container(self.column_viewer, "Columns"))

    # Add a dock to the MainWindow widget
    def add_view(self, widget: QtWidgets.QWidget, title: str):
        dock = DockWidget(title, self.pgdf.name)
        dock.setAllowedAreas(Qt.AllDockWidgetAreas)

        frame = QtWidgets.QFrame()
        frame.setFrameStyle(frame.Box | frame.Raised)
        frame.setLineWidth(2)
        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(widget)
        frame.setLayout(layout)
        dock.setWidget(frame)

        if len(self.docks) > 0:
            self.main_window.tabifyDockWidget(self.docks[0], dock)
            # Keep the first tab active by default
            self.docks[0].raise_()
        else:
            self.main_window.addDockWidget(Qt.LeftDockWidgetArea, dock)

        self.docks.append(dock)
        return dock


class ColumnArranger(ColumnViewer):

    def __init__(self, pgdf: PandasGuiDataFrameStore):
        super().__init__(pgdf)
        pgdf = PandasGuiDataFrameStore.cast(pgdf)
        self.pgdf = pgdf

        self.tree.setAcceptDrops(True)
        self.tree.setDragEnabled(True)
        self.tree.setDefaultDropAction(Qt.MoveAction)
        self.tree.setHeaderLabels(['Name'])
        self.refresh()

        self.setContextMenuPolicy(Qt.CustomContextMenu)

        self.tree.onDropSignal.connect(self.columns_rearranged)
        self.tree.itemDoubleClicked.connect(self.on_double_click)

        self.tree.mouseReleaseEventSignal.connect(self.on_mouseReleaseEvent)

    def on_double_click(self, item: QtWidgets.QTreeWidgetItem):
        ix = self.tree.indexOfTopLevelItem(item)
        self.pgdf.dataframe_viewer.scroll_to_column(ix)

    def columns_rearranged(self):
        items = [self.tree.topLevelItem(x).text(0) for x in range(self.tree.topLevelItemCount())]
        self.pgdf.reorder_columns(items)

    def sizeHint(self) -> QtCore.QSize:
        return QtCore.QSize(200, super().sizeHint().height())

    def on_mouseReleaseEvent(self, event):
        # TODO - Add context menu to move columns to front and end
        return
        if event.button() == Qt.RightButton:
            pos = event.pos()
            item = self.tree.itemAt(pos)
            ix = self.tree.indexAt(pos)
            ix_list = [self.tree.indexOfTopLevelItem(x) for x in self.tree.selectedItems()]
            if item:
                menu = QtWidgets.QMenu(self.tree)

                action1 = QtWidgets.QAction("Move To Front")
                action1.triggered.connect(lambda: None)

                action2 = QtWidgets.QAction("Move To End")
                action2.triggered.connect(lambda: None)

                for action in [action1, action2]:
                    menu.addAction(action)

                menu.exec_(QtGui.QCursor().pos())

        super().mouseReleaseEvent(event)


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)

    from pandasgui.datasets import pokemon

    # Create and show widget
    dfe = DataFrameExplorer(pokemon)
    dfe.show()

    sys.exit(app.exec_())



from PyQt5 import QtWidgets


class CollapsiblePanel(QtWidgets.QGroupBox):
    def __init__(self, title="", parent=None):
        super().__init__(parent)
        self.setCheckable(True)
        self.setTitle(title)

        self.contents_widget = QtWidgets.QWidget(self)
        self.toggled.connect(self.on_toggled)
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.contents_widget)

    def setContent(self, widget) -> None:
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(widget)
        self.contents_widget.setLayout(layout)
        super().setLayout(self.layout)

    def on_toggled(self, checked: bool = ...) -> None:
        if checked:
            self.contents_widget.show()
        else:
            self.contents_widget.hide()


if __name__ == "__main__":
    app = QtWidgets.QApplication([])
    x = CollapsiblePanel("title")
    w = QtWidgets.QLineEdit("asdf")
    x.setContent(w)
    x.show()
    app.exec_()



from PyQt5 import QtGui, QtWidgets
from PyQt5.QtCore import pyqtSignal


class QTreeWidget(QtWidgets.QTreeWidget):
    onDropSignal = pyqtSignal()

    # Autosize columns
    def showEvent(self, event: QtGui.QShowEvent):
        self.autosize_columns()
        event.accept()

    def autosize_columns(self):
        for i in range(self.columnCount()):
            self.resizeColumnToContents(i)

    def dropEvent(self, event: QtGui.QDropEvent) -> None:
        super().dropEvent(event)
        self.onDropSignal.emit()

class QTreeWidgetItem(QtWidgets.QTreeWidgetItem):
    def __lt__(self, otherItem):
        column = self.treeWidget().sortColumn()

        try:
            return float(self.text(column)) < float(otherItem.text(column))
        except ValueError:
            return self.text(column) < otherItem.text(column)



'''
Menu that appears on right clicking a column header. Contains options for modifying DataFrame columns
'''

import sys

from PyQt5 import QtCore, QtGui, QtWidgets

from pandasgui.store import PandasGuiDataFrameStore


class ColumnMenu(QtWidgets.QMenu):
    def __init__(self, pgdf: PandasGuiDataFrameStore, column_ix: int, parent=None):
        super().__init__(parent)

        self.pgdf = pgdf
        self.column_ix = column_ix
        df = self.pgdf.df_unfiltered

        ########################
        # Info
        label = QtWidgets.QLabel(self.pgdf.df_unfiltered.columns[column_ix])
        self.add_widget(label)

        ########################
        # Sorting

        def select_button():
            self.sort_b1.setDown(self.pgdf.sort_state == 'Asc' and self.pgdf.sorted_column_ix == column_ix)
            self.sort_b2.setDown(self.pgdf.sort_state == 'Desc' and self.pgdf.sorted_column_ix == column_ix)
            self.sort_b3.setDown(self.pgdf.sort_state == 'None' or self.pgdf.sorted_column_ix != column_ix)

        self.sort_b1 = QtWidgets.QPushButton("Asc")
        self.sort_b1.clicked.connect(lambda: [self.pgdf.sort_column(self.column_ix, "Asc"),
                                              select_button()])

        self.sort_b2 = QtWidgets.QPushButton("Desc")
        self.sort_b2.clicked.connect(lambda: [self.pgdf.sort_column(self.column_ix, "Desc"),
                                              select_button()])

        self.sort_b3 = QtWidgets.QPushButton("None")
        self.sort_b3.clicked.connect(lambda: [self.pgdf.sort_column(self.column_ix, "None"),
                                              select_button()])

        select_button()

        sort_control = QtWidgets.QWidget()
        sort_control_layout = QtWidgets.QHBoxLayout()
        sort_control_layout.setSpacing(0)
        sort_control_layout.setContentsMargins(0, 0, 0, 0)
        sort_control.setLayout(sort_control_layout)
        [sort_control_layout.addWidget(w) for w in [self.sort_b1, self.sort_b2, self.sort_b3]]

        self.add_widget(sort_control)

        ########################
        # Move

        col_name = self.pgdf.df.columns[column_ix]
        self.move_b1 = QtWidgets.QPushButton("<<")
        self.move_b1.clicked.connect(lambda: [self.pgdf.move_column(column_ix, 0),
                                              self.close(), self.pgdf.dataframe_viewer.show_column_menu(col_name)])
        self.move_b2 = QtWidgets.QPushButton("<")
        self.move_b2.clicked.connect(lambda: [self.pgdf.move_column(column_ix, column_ix - 1),
                                              self.close(), self.pgdf.dataframe_viewer.show_column_menu(col_name)])
        self.move_b3 = QtWidgets.QPushButton(">")
        self.move_b3.clicked.connect(lambda: [self.pgdf.move_column(column_ix, column_ix + 1),
                                              self.close(), self.pgdf.dataframe_viewer.show_column_menu(col_name)])
        self.move_b4 = QtWidgets.QPushButton(">>")
        self.move_b4.clicked.connect(lambda: [self.pgdf.move_column(column_ix, len(df.columns)),
                                              self.close(), self.pgdf.dataframe_viewer.show_column_menu(col_name)])

        move_control = QtWidgets.QWidget()
        move_control_layout = QtWidgets.QHBoxLayout()
        move_control_layout.setSpacing(0)
        move_control_layout.setContentsMargins(0, 0, 0, 0)
        move_control.setLayout(move_control_layout)
        [move_control_layout.addWidget(w) for w in [self.move_b1, self.move_b2, self.move_b3, self.move_b4]]

        self.add_widget(move_control)

        ########################
        # Delete

        button = QtWidgets.QPushButton("Delete Column")
        button.clicked.connect(lambda: [self.pgdf.delete_column(column_ix), self.close()])
        self.add_widget(button)

        ########################
        # Parse Date

        button = QtWidgets.QPushButton("Parse Date")
        button.clicked.connect(lambda: [self.pgdf.parse_date(column_ix), self.close()])
        self.add_widget(button)

        ########################
        # Data Type
        col_type = self.pgdf.df_unfiltered.dtypes[column_ix]
        types = list(dict.fromkeys([str(col_type)] + ['float', 'bool', 'category', 'str']))
        combo = QtWidgets.QComboBox()
        combo.addItems(types)
        combo.setCurrentText(str(col_type))
        combo.currentTextChanged.connect(lambda x: [self.pgdf.change_column_type(column_ix, x)])
        self.add_widget(combo)

        ########################
        # Coloring
        self.add_action("Color by None",
                        lambda: [setattr(self.pgdf.dataframe_viewer, 'color_mode', None),
                                 self.pgdf.dataframe_viewer.refresh_ui()]
                        )

        self.add_action("Color by columns",
                        lambda: [setattr(self.pgdf.dataframe_viewer, 'color_mode', 'column'),
                                 self.pgdf.dataframe_viewer.refresh_ui()]
                        )

        self.add_action("Color by rows",
                        lambda: [setattr(self.pgdf.dataframe_viewer, 'color_mode', 'row'),
                                 self.pgdf.dataframe_viewer.refresh_ui()]
                        )

        self.add_action("Color by all",
                        lambda: [setattr(self.pgdf.dataframe_viewer, 'color_mode', 'all'),
                                 self.pgdf.dataframe_viewer.refresh_ui()]
                        )

    def add_action(self, text, function):
        action = QtWidgets.QAction(text, self)
        action.triggered.connect(function)
        self.addAction(action)

    def add_widget(self, widget):
        # https://stackoverflow.com/questions/55086498/highlighting-custom-qwidgetaction-on-hover
        widget.setMouseTracking(True)

        custom_action = QtWidgets.QWidgetAction(self)
        widget.setStyleSheet(widget.styleSheet() + "margin: 5px;")
        custom_action.setDefaultWidget(widget)
        self.addAction(custom_action)

    def show_menu(self, point):
        self.move(point)
        self.show()


if __name__ == "__main__":
    # Create a QtWidgets.QApplication instance or use the existing one if it exists
    app = QtWidgets.QApplication(sys.argv)
    from pandasgui.datasets import pokemon
    from pandasgui import show

    # gui = show(pokemon)

    w = ColumnMenu(pokemon, 0)
    app.exec_()



import pandas as pd
import numpy as np
import time
import os
from PyQt5 import QtCore, QtWidgets


def generate_string_data(rows, cols, length=5):
    return pd.DataFrame(np.array([pd.util.testing.rands_array(length, cols) for x in range(rows)]))


def generate_int_data(rows, cols):
    return pd.DataFrame(np.random.randint(0, 100, size=(rows, cols)))


def generate_data():
    small_int_path = os.path.join("small_int.csv")
    large_int_path = "large_int.csv"
    small_str_path = "small_str.csv"
    large_str_path = "large_str.csv"

    generate_int_data(5000, 5).to_csv(small_int_path)
    generate_int_data(5000000, 50).to_csv(large_int_path)
    generate_string_data(5000, 5).to_csv(small_str_path)
    generate_string_data(5000000, 50).to_csv(large_str_path)


def test_memory_leak():
    from pandasgui import show

    def getObjects(cls):
        import gc
        objects = []
        for obj in gc.get_objects():
            if isinstance(obj, cls):
                objects.append(obj)
        return objects

    # string_data = generate_string_data(rows, cols)
    int_data = pd.read_csv(large_int_path)

    # %%

    for i in range(10):
        df = pd.DataFrame(int_data)

        print(len(getObjects(QtCore.QObject)), 'QObjects')
        start = time.time()
        gui = show(df, settings={'block': False})
        gui.close()
        end = time.time()
        print(end - start)


def except_hook(cls, exception, traceback):
    sys.__excepthook__(cls, exception, traceback)


import sys

sys.excepthook = except_hook


def test_webengine_import():
    app = QtWidgets.QApplication([])
    from pandasgui.widgets.figure_viewer import FigureViewer
    import plotly.graph_objs as go

    fig = go.Figure()
    fig.add_scatter(
        x=np.random.rand(100),
        y=np.random.rand(100),
        mode="markers",
        marker={
            "size": 30,
            "color": np.random.rand(100),
            "opacity": 0.6,
            "colorscale": "Viridis",
        },
    )

    pv = FigureViewer(fig)
    pv.show()


def test_inputs():
    import pandas as pd
    import numpy as np
    from pandasgui import show
    iterables = [["bar", 57, 57], ["one", 23]]
    ix = pd.MultiIndex.from_product(iterables, names=["first", "second"])
    df = pd.DataFrame(np.random.randn(6, 6), index=ix[:6], columns=ix[:6])

    df2 = pd.DataFrame({'a': [1, 2, 3], 'b': [4, 5, 6],
                        2: [1, 2, 3], 'c': [4, 5, 6], }).rename(columns={'c': 2})

    from pandasgui.datasets import all_datasets
    show(df, df2, **all_datasets, settings={'block': False})


def test_code_history():
    import pandas as pd
    from pandasgui import show
    from pandasgui.datasets import pokemon
    pokemon = pokemon.head(10)[['Name', 'Attack', 'Defense', 'Generation', 'HP', 'Legendary']]
    gui = show(pokemon, settings={'block': False})
    pgdf = gui.store.data['pokemon']

    pgdf.edit_data(6, 3, "999")
    pgdf.sort_column(5)
    pgdf.edit_data(3, 4, "5")
    pgdf.paste_data(0, 1, pd.DataFrame({0: {0: 1000, 1: 1001}, 1: {0: 1002, 1: 1003}, 2: {0: 1004, 1: 1005}}))
    pgdf.add_filter('HP > 50')
    pgdf.sort_column(4)

    code = pgdf.code_export()
    # https://stackoverflow.com/a/52217741/3620725
    df = pokemon
    namespace = {'df': df}
    exec(code, namespace)
    df = namespace['df']

    assert (df.fillna('NULL').equals(gui.get_dataframes('pokemon').fillna('NULL')))


def test_json():
    import requests
    from pandasgui import show
    from pandasgui.datasets import all_datasets
    comments = requests.get('https://jsonplaceholder.typicode.com/comments').json()
    photos = requests.get('https://jsonplaceholder.typicode.com/photos').json()

    gui = show(comments, photos, **all_datasets, settings={'block': False})


test_json()
test_inputs()
test_code_history()
# test_webengine_import()

QtWidgets.QApplication
# iterables = [["bar", "baz", "baz"], ["one", "two"]]
# ix = pd.MultiIndex.from_product(iterables, names=["first", "second"])
# df = pd.DataFrame(np.random.randn(6, 6), index=ix[:6], columns=ix[:6])



import datetime
from collections import defaultdict

import plotly.express as px
from pandas import DataFrame
from typing_extensions import Literal
from plotly.graph_objs import Figure
import plotly.graph_objects as go

from typing import List
import pandas as pd
from pandasgui.store import SETTINGS_STORE, PandasGuiDataFrameStore


class HiddenArg(type):
    pass


class ColumnName(str):
    pass


class ColumnNameList(List[str]):
    pass


class OtherDataFrame(DataFrame):
    pass


# ============================================================================ #
# Graphing

def histogram(data_frame: DataFrame,
              x: ColumnName = None,
              color: ColumnName = None,
              facet_row: ColumnName = None,
              facet_col: ColumnName = None,
              marginal: Literal[None, 'rug', 'box', 'violin'] = None,
              cumulative: bool = False,
              **kwargs) -> Figure:
    fig = px.histogram(data_frame=data_frame,
                       x=x,
                       color=color,
                       facet_row=facet_row,
                       facet_col=facet_col,
                       marginal=marginal,
                       cumulative=cumulative,
                       **kwargs)

    return fig


def scatter(data_frame: DataFrame,
            y: ColumnName = None,
            x: ColumnName = None,
            color: ColumnName = None,
            facet_row: ColumnName = None,
            facet_col: ColumnName = None,
            symbol: ColumnName = None,
            size: ColumnName = None,
            trendline: Literal[None, 'ols', 'lowess'] = None,
            marginal: Literal[None, 'histogram', 'rug', 'box', 'violin'] = None,
            **kwargs) -> Figure:
    if size is not None:
        data_frame = data_frame.dropna(subset=[size])
    fig = px.scatter(data_frame=data_frame,
                     x=x,
                     y=y,
                     color=color,
                     symbol=symbol,
                     size=size,
                     trendline=trendline,
                     marginal_x=marginal,
                     marginal_y=marginal,
                     facet_row=facet_row,
                     facet_col=facet_col,

                     render_mode=SETTINGS_STORE.render_mode.value,

                     **kwargs)
    fig.update()

    return fig


def line(data_frame: DataFrame,
         y: ColumnName = None,
         x: ColumnName = None,
         color: ColumnName = None,
         facet_row: ColumnName = None,
         facet_col: ColumnName = None,
         # Custom args
         aggregation: Literal['mean', 'median', 'min', 'max', 'sum', None] = 'mean',
         **kwargs) -> Figure:
    # Create list of key columns
    key_cols = [val for val in [x, color, facet_row, facet_col] if val is not None]
    if key_cols != []:
        if aggregation is not None:
            data_frame = data_frame.groupby(key_cols).agg(aggregation).reset_index()
        else:
            # Only need to sort here because aggregation already sorts
            data_frame = data_frame.sort_values(key_cols)

    fig = px.line(data_frame=data_frame,
                  x=x,
                  y=y,
                  color=color,
                  facet_row=facet_row,
                  facet_col=facet_col,

                  render_mode=SETTINGS_STORE.render_mode.value,

                  **kwargs)

    # Don't want gaps in line where there are NaN datapoints
    fig.update_traces(connectgaps=True)

    return fig


def bar(data_frame: DataFrame,
        y: ColumnName = None,
        x: ColumnName = None,
        color: ColumnName = None,
        facet_row: ColumnName = None,
        facet_col: ColumnName = None,
        # Custom args
        aggregation: Literal['mean', 'median', 'min', 'max', 'sum', None] = 'mean',
        **kwargs) -> Figure:
    # Create list of key columns
    key_cols = [val for val in [x, color, facet_row, facet_col] if val is not None]
    if key_cols != []:
        if aggregation is not None:
            data_frame = data_frame.groupby(key_cols).agg(aggregation).reset_index()
        else:
            # Only need to sort here because aggregation already sorts
            data_frame = data_frame.sort_values(key_cols)

    fig = px.bar(data_frame=data_frame,
                 x=x,
                 y=y,
                 color=color,
                 facet_row=facet_row,
                 facet_col=facet_col,
                 **kwargs)

    return fig


def box(data_frame: DataFrame,
        y: ColumnName = None,
        x: ColumnName = None,
        color: ColumnName = None,
        facet_row: ColumnName = None,
        facet_col: ColumnName = None,
        **kwargs) -> Figure:
    fig = px.box(data_frame=data_frame,
                 x=x,
                 y=y,
                 color=color,
                 facet_row=facet_row,
                 facet_col=facet_col,

                 **kwargs)

    return fig


def violin(data_frame: DataFrame,
           y: ColumnName = None,
           x: ColumnName = None,
           color: ColumnName = None,
           facet_row: ColumnName = None,
           facet_col: ColumnName = None,
           **kwargs) -> Figure:
    fig = px.violin(data_frame=data_frame,
                    x=x,
                    y=y,
                    color=color,
                    facet_row=facet_row,
                    facet_col=facet_col,

                    **kwargs)

    return fig


def density_heatmap(data_frame: DataFrame,
                    y: ColumnName = None,
                    x: ColumnName = None,
                    z: ColumnName = None,
                    facet_row: ColumnName = None,
                    facet_col: ColumnName = None,
                    **kwargs) -> Figure:
    fig = px.density_heatmap(data_frame=data_frame,
                             x=x,
                             y=y,
                             z=z,
                             facet_row=facet_row,
                             facet_col=facet_col,

                             **kwargs)

    return fig


def density_contour(data_frame: DataFrame,
                    y: ColumnName = None,
                    x: ColumnName = None,
                    z: ColumnName = None,
                    facet_row: ColumnName = None,
                    facet_col: ColumnName = None,
                    **kwargs) -> Figure:
    fig = px.density_contour(data_frame=data_frame,
                             x=x,
                             y=y,
                             z=z,
                             facet_row=facet_row,
                             facet_col=facet_col,

                             **kwargs)

    fig.update_traces(contours_coloring="fill", contours_showlabels=True)

    return fig


def pie(data_frame: DataFrame,
        names: ColumnName = None,
        values: ColumnName = None,
        color: ColumnName = None,
        facet_row: ColumnName = None,
        facet_col: ColumnName = None,
        **kwargs) -> Figure:
    if facet_row is not None or facet_col is not None:
        raise NotImplementedError

    fig = px.pie(data_frame=data_frame,
                 names=names,
                 values=values,
                 color=color,
                 **kwargs)

    return fig


def scatter_matrix(data_frame: DataFrame,
                   dimensions: ColumnNameList = None,
                   color: ColumnName = None,
                   **kwargs) -> Figure:
    fig = px.scatter_matrix(data_frame=data_frame,
                            dimensions=dimensions,
                            color=color,

                            **kwargs)

    return fig


def scatter_3d(data_frame: DataFrame,
               y: ColumnName = None,
               x: ColumnName = None,
               z: ColumnName = None,
               color: ColumnName = None,
               **kwargs) -> Figure:
    fig = px.scatter_3d(data_frame=data_frame,
                        x=x,
                        y=y,
                        z=z,
                        color=color,
                        **kwargs)

    return fig


def candlestick(data_frame: DataFrame,
                x: ColumnName = None,
                open: ColumnName = None,
                high: ColumnName = None,
                low: ColumnName = None,
                close: ColumnName = None,
                **kwargs) -> Figure:
    fig = go.Figure(data=[go.Candlestick(x=data_frame[x],
                                         open=data_frame[open],
                                         high=data_frame[high],
                                         low=data_frame[low],
                                         close=data_frame[close],
                                         )])

    return fig


def word_cloud(data_frame: DataFrame,
               words: ColumnNameList = None,
               **kwargs) -> Figure:
    from wordcloud import WordCloud

    # TODO look at this https://github.com/PrashantSaikia/Wordcloud-in-Plotly/blob/master/plotly_wordcloud.py

    if type(words) == str:
        words = [words]
    text = ' '.join(pd.concat([data_frame[x].dropna().astype(str) for x in words]))
    wc = WordCloud(scale=2, collocations=False).generate(text)
    fig = px.imshow(wc)

    return fig


# ============================================================================ #
# Reshaping


def pivot(data_frame: DataFrame,
          index: ColumnName = None,
          columns: ColumnName = None,
          values: ColumnName = None,
          # https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html#groupby-aggregate-named
          aggfunc: Literal['count',
                           'mean',
                           'median',
                           'mode',
                           'min',
                           'max',
                           'sum',
                           'nunique'] = 'mean'
          ):
    df = data_frame.pivot_table(index=index,
                                columns=columns,
                                values=values,
                                aggfunc=aggfunc)
    return df


def melt(df: DataFrame,
         id_vars: ColumnName = None,
         value_vars: ColumnName = None,
         ):
    df = df.melt(id_vars=id_vars,
                 value_vars=value_vars)
    return df


def merge(data_frame: DataFrame,
          other_dataframe: OtherDataFrame,
          how: Literal['left', 'right', 'outer', 'inner'] = 'inner',
          left_on: ColumnNameList = None,
          right_on: ColumnNameList = None,

          validate: Literal['one_to_one', 'one_to_many', 'many_to_one', 'many_to_many'] = None,
          ):
    df = data_frame.merge(right=other_dataframe,
                          how=how,
                          left_on=left_on,
                          right_on=right_on,
                          validate=validate, )
    return df


def concat(data_frame: DataFrame,
           other_dataframe: OtherDataFrame,
           axis: Literal['0 (rows)', '1 (columns)'] = '0 (rows)',
           ):
    df = pd.concat([data_frame, other_dataframe],
                   axis=int(axis[0])
                   )
    return df


# ============================================================================ #
# Utility functions


def generate_title(pgdf: PandasGuiDataFrameStore, chart_type, kwargs):
    """template variable replacement.

    Besides all dragger selections (x, y, z, color etc), and all custom kwargs, extra template variables are:

          date: current datetime
          filters: active filter expressions, comma delimited
          title_x: x minus units and with log scale if selected
          title_y: y minus units and with log scale if selected
          title_z: z minus units and with log scale if selected
          title_dimensions: dimensions list minus units
          title_columns: columns list minus units
          title_trendline: trendline description
          vs: when doing a title with x vs y, use {x}{vs}{y}
          over_by: when doing a title y over x, use {y}{over_by}{x}. Preferred. for distributions, will use "by"
          name: dataframe name
          total: total number of observations
          subset: observations with active filters applied
          selection: ready to use string representing {subset} observations of {total}
          groupings: groupings tied to Legend and not on legend: marker_symbol, line_group, size etc
    """

    def remove_units(label):
        "we do not want to repeat units in the title. It is assumed they are in parenthesis"
        if type(label) == list and len(label) == 0:
            return label
        elif type(label) == list and len(label) == 1:
            return remove_units(label[0])
        elif type(label) == list and len(label) > 1:
            return [remove_units(val) for val in label]
        try:
            return label[:label.rindex("(")] if label[-1] == ")" else label
        except (AttributeError, IndexError, TypeError, ValueError):
            return label

    def axis_title_labels(kwargs):
        """axis_title_labels.

        handling of x, y, z and dimensions, columns / orientation and log scales for all charts.

        Args:
            kwargs: PX keywords related to axes and also dimensions and columns facets.
        """
        orientation = kwargs.get("orientation")
        log_x = kwargs.get("log_x", False)
        log_y = kwargs.get("log_y", False)
        log_z = kwargs.get("log_z", False)

        title_log_x = "log " if log_x else ""
        title_log_y = "log " if log_y else ""
        title_log_z = "log " if log_z else ""

        x = remove_units(kwargs.get("x"))
        y = remove_units(kwargs.get("y", ""))
        z = remove_units(kwargs.get("z", ""))
        dimensions = remove_units(kwargs.get("dimensions", ""))
        columns = remove_units(kwargs.get("columns", ""))

        if orientation == "h":
            x, y = y, x
            title_log_x, title_log_y = title_log_y, title_log_x
        if x is None:
            opt_x = x  # wordcloud, pie etc have no x
        elif type(x) == list:
            opt_x = [title_log_x + val for val in x]
        else:
            opt_x = title_log_x + x
        if type(y) == list:
            opt_y = [title_log_y + val for val in y]
        else:
            opt_y = title_log_y + y
        if z is None:
            z = ""
        return opt_x, opt_y, title_log_z + z, dimensions, columns

    try:
        name = pgdf.name
        pgdf_filters = pgdf.filters
        df_unfiltered = pgdf.df_unfiltered
        df = pgdf.df
    except AttributeError:  # plain dataframe
        name = kwargs.pop("name", "untitled")
        pgdf_filters = pgdf.filters
        df_unfiltered = pgdf.df_unfiltered
        df = pgdf.df

    today = datetime.datetime.now()
    x, y, z, dimensions, columns = axis_title_labels(kwargs)
    title = kwargs.get("kwargs", {}).get("title", None)  # not explicitly kwargs, from locals()

    # Histograms default to count for aggregation
    histfunc = ""
    over_by = " over "
    vs = " vs "
    color = kwargs.get("color")
    symbol = kwargs.get("symbol")
    aggregation = kwargs.get("aggregation", None)
    title_trendline = kwargs.get("trendline", None)
    if title_trendline:
        title_trendline = f"trend lines are <i>{title_trendline}</i>"
    else:
        title_trendline = ""

    if chart_type == "histogram":
        histfunc = kwargs.get("histfunc", "sum" if y else "count")
        if x is None and y:
            y = f"{y} {histfunc}"
    elif chart_type in ("box", "violin"):
        histfunc = "distribution"
        over_by = " by "
        if x is None and y:
            y = f"{y} {histfunc}"
    elif chart_type == "bar":
        if x is None:
            x = "index"
        over_by = " by "
        if y == "":
            y = "count"
            if color is None and aggregation:
                over_by = " of "
        else:
            if aggregation:
                func = "average"
            else:
                func = "sum"
            y = f"{func} of {y}"
    elif chart_type == "density_heatmap":
        histfunc = kwargs.get("histfunc", "sum")
        if y and z:
            y, z = f"binned {histfunc} of {z} for ", y
        elif y:
            y = f"binned count of {y}"
        histfunc = ""
    elif chart_type == "density_contour":
        histfunc = kwargs.get("histfunc", "count")
        estimation = "estimated density "
        if y and z:
            y, z = f"estimated {histfunc} density of {z} for ", y
        elif y:
            y = f"estimated count density of {y}"
        histfunc = ""
    elif chart_type == "scatter":
        if x is None:
            x = "index"
    elif chart_type == "scatter_3d":
        if y and z:
            # need to separate them
            z = ", " + z
    elif chart_type in ("word_cloud", "scatter_matrix", "pie"):
        x = ""  # else string will evaluate to None
    elif chart_type == "line":
        if x is None:
            x = "index"
        if kwargs.get("aggregation", None):
            over_by = " by "
            if y:
                y = f"{aggregation} of {y}"

    # filters
    filters = ','.join([filter.expr for filter in pgdf_filters if filter.enabled and filter.failed == False])
    if filters != "":
        filters = "Filters: " + filters
    total = df_unfiltered.shape[0]

    subset = df.shape[0]
    selection = ""
    groupings = ""
    sep = ""
    showlegend = kwargs.get("showlegend", True)

    # over / by
    over_by = f" {histfunc}{over_by}" if x else ""
    vs = f"{vs} {histfunc}" if y else ""

    # Groupings in Legend
    if color or symbol:
        if showlegend:
            groupings += "Legend"
        else:
            if color:
                groupings += f"{sep}color={color}"
                sep = ", "
            if symbol:
                groupings += f"{sep}symbol={symbol}"
        sep = ", "

    # this one shows on the legend, but is not labeled
    if kwargs.get("marker_symbol"):
        groupings += f"{sep}marker={kwargs['marker_symbol']}"

    # next three don't show in the plotly legend, so we need to explicitly add them
    if "line_group" in kwargs.keys() and kwargs['line_group'] is not None:
        groupings += f"{sep}line_group={kwargs['line_group']}"
    if "size" in kwargs.keys() and kwargs['size'] is not None:
        groupings += f"{sep}size={kwargs['size']}"
    if "text" in kwargs.keys() and kwargs['text'] is not None:
        groupings += f"{sep}text={kwargs['text']}"
    if groupings != "":
        groupings = "Grouped by " + groupings
        if filters != "":
            groupings += " - "

    if subset != total:
        selection = f"({subset} obs. of {total}) "
    elif aggregation:
        selection = "({groupby_obs} " + f"derived obs. from {total})"
    else:
        selection = f"({total} obs.)" if chart_type == "line" else "(all observations)"

    title_format = kwargs.get("title_template", SETTINGS_STORE['title_format'].value)
    title_rendered = title_format. \
        format_map(defaultdict(str,
                               date=today,
                               filters=filters,
                               title_x=x,
                               title_y=y,
                               title_z=z,
                               title_dimensions=dimensions,
                               title_columns=columns,
                               title_trendline=title_trendline,
                               vs=vs,
                               over_by=over_by,
                               name=name,
                               total=total,
                               subset=subset,
                               selection=selection,
                               groupings=groupings,
                               **kwargs))
    if title:
        title = title.replace("{title}", title_rendered)  # to include auto title in custom kwargs title
    else:
        title = title_rendered
    return title



"""A ComboBox widget with a search box to filter the contents"""

# https://stackoverflow.com/a/7693234/3620725

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import Qt


class SearchableComboBox(QtWidgets.QComboBox):
    def __init__(self, string_list, parent=None):
        super(SearchableComboBox, self).__init__(parent)

        self.setFocusPolicy(Qt.StrongFocus)
        self.setEditable(True)

        # add a filter model to filter matching items
        self.pFilterModel = QtCore.QSortFilterProxyModel(self)
        self.pFilterModel.setFilterCaseSensitivity(Qt.CaseInsensitive)
        self.pFilterModel.setSourceModel(self.model())

        # add a completer, which uses the filter model
        self.completer = QtWidgets.QCompleter(self.pFilterModel, self)
        # always show all (filtered) completions
        self.completer.setCompletionMode(QtWidgets.QCompleter.UnfilteredPopupCompletion)
        self.setCompleter(self.completer)

        # connect signals
        self.lineEdit().textEdited.connect(self.pFilterModel.setFilterFixedString)
        self.completer.activated.connect(self.on_completer_activated)

        # either fill the standard model of the combobox
        self.addItems(string_list)

    # on selection of an item from the completer, select the corresponding item from combobox
    def on_completer_activated(self, text):
        if text:
            index = self.findText(text)
            self.setCurrentIndex(index)
            self.activated[str].emit(self.itemText(index))

    # on model change, update the models of the filter and completer as well
    def setModel(self, model):
        super(SearchableComboBox, self).setModel(model)
        self.pFilterModel.setSourceModel(model)
        self.completer.setModel(self.pFilterModel)

    # on model column change, update the model column of the filter and completer as well
    def setModelColumn(self, column):
        self.completer.setCompletionColumn(column)
        self.pFilterModel.setFilterKeyColumn(column)
        super(SearchableComboBox, self).setModelColumn(column)


if __name__ == "__main__":
    import sys
    from PyQt5.QtWidgets import QApplication

    app = QApplication(sys.argv)
    string_list = ["hola muchachos", "adios amigos", "hello world", "good bye"]
    combo = SearchableComboBox(string_list)

    combo.resize(300, 40)
    combo.show()

    sys.exit(app.exec_())



import sys

from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QMouseEvent

from pandasgui.store import PandasGuiDataFrameStore
from pandasgui.utility import nunique
from pandasgui.widgets import base_widgets


class FlatDraggableTree(base_widgets.QTreeWidget):
    mouseReleaseEventSignal = pyqtSignal(QMouseEvent)

    def __init__(self):
        super().__init__()

        self.header().setStretchLastSection(False)
        self.setDragDropMode(self.InternalMove)
        self.setSelectionMode(self.ExtendedSelection)
        self.setSelectionBehavior(self.SelectRows)
        self.setRootIsDecorated(False)

        self.expandAll()
        self.apply_tree_settings()

    def showEvent(self, event: QtGui.QShowEvent):
        for i in range(self.columnCount()):
            self.resizeColumnToContents(i)
        event.accept()

    def rowsInserted(self, parent: QtCore.QModelIndex, start: int, end: int):
        super().rowsInserted(parent, start, end)
        self.expandAll()

    def sizeHint(self) -> QtCore.QSize:
        self.header().setStretchLastSection(False)
        width = 5 + sum([self.columnWidth(i) for i in range(self.columnCount())])
        self.header().setStretchLastSection(True)

        return QtCore.QSize(width, super().sizeHint().height())

    def dragEnterEvent(self, e: QtGui.QDragEnterEvent) -> None:
        self.apply_tree_settings()
        super().dragEnterEvent(e)

    def apply_tree_settings(self):
        root = self.invisibleRootItem()
        root.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDropEnabled)

        for i in range(root.childCount()):
            child = root.child(i)
            child.setExpanded(True)

            child.setFlags(Qt.ItemIsEnabled | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled & ~Qt.ItemIsDropEnabled)

    def mimeData(self, indexes):
        mimedata = super().mimeData(indexes)
        if indexes:
            mimedata.setText(indexes[0].text(0))
        return mimedata

    def mouseReleaseEvent(self, event):
        self.mouseReleaseEventSignal.emit(event)
        super().mouseReleaseEvent(event)


class ColumnViewer(QtWidgets.QWidget):
    def __init__(self, pgdf: PandasGuiDataFrameStore):
        super().__init__()

        self.tree: FlatDraggableTree = FlatDraggableTree()
        self.tree.setHeaderLabels(['Name'])
        self.tree.setDragEnabled(True)
        self.tree.setDefaultDropAction(Qt.CopyAction)

        # Search box
        self.search_bar = QtWidgets.QLineEdit()
        self.search_bar.textChanged.connect(self.filter)

        self.source_tree_layout = QtWidgets.QVBoxLayout()
        self.source_tree_layout.addWidget(self.search_bar)
        self.source_tree_layout.addWidget(self.tree)
        self.setLayout(self.source_tree_layout)

        self.update_df(pgdf)

    def update_df(self, df):
        pgdf = PandasGuiDataFrameStore.cast(df)

        self.pgdf = pgdf
        self.refresh()

    def refresh(self):
        sources = self.pgdf.df_unfiltered.columns
        source_nunique = nunique(self.pgdf.df_unfiltered)
        source_types = self.pgdf.df_unfiltered.dtypes.values.astype(str)

        # Ensure no duplicates
        assert (len(sources) == len(set(sources)))
        assert (len(sources) == len(source_nunique))
        assert (len(sources) == len(source_types))

        self.tree.clear()
        for i in range(len(sources)):
            item = base_widgets.QTreeWidgetItem(self.tree,
                                                [str(sources[i]), str(source_nunique[i]), str(source_types[i])])

        # Depends on Search Box and Source list
        self.filter()
        self.tree.apply_tree_settings()

    def filter(self):
        root = self.tree.invisibleRootItem()
        for i in range(root.childCount()):
            child = root.child(i)
            child.setHidden(True)

        items = self.tree.findItems(f".*{self.search_bar.text()}.*",
                                    Qt.MatchRegExp | Qt.MatchRecursive)
        for item in items:
            item.setHidden(False)


class SortableColumnViewer(ColumnViewer):

    def __init__(self, pgdf: PandasGuiDataFrameStore):
        super().__init__(pgdf)
        self.tree.setSortingEnabled(True)


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)

    from pandasgui.datasets import pokemon

    x = ColumnViewer(pokemon)
    x.show()

    sys.exit(app.exec_())



from PyQt5 import QtWidgets


class Container(QtWidgets.QGroupBox):
    def __init__(self, widget, title):
        super().__init__()

        self.setTitle(title)
        self.layout = QtWidgets.QVBoxLayout()
        self.setLayout(self.layout)
        self.layout.addWidget(widget)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.setFlat(True)



import logging
import pandas as pd
from PyQt5 import QtWidgets
from typing import List, Union
import sys
import inspect
from collections import OrderedDict

logger = logging.getLogger(__name__)


class DotDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __init__(self, dct):
        for key, value in dct.items():
            if hasattr(value, "keys"):
                value = DotDict(value)
            self[key] = value


class SlicableOrderedDict(OrderedDict):
    def __getitem__(self, k):
        if not isinstance(k, slice):
            return OrderedDict.__getitem__(self, k)
        x = SlicableOrderedDict()
        for idx, key in enumerate(self.keys()):
            if k.start <= idx < k.stop:
                x[key] = self[key]
        return x


def throttle(wait):
    """
        https://gist.github.com/walkermatt/2871026
        Decorator that will postpone a functions
        execution until after wait seconds
        have elapsed since the last time it was invoked. """
    from threading import Timer
    def decorator(fn):
        def throttled(*args, **kwargs):
            def call_it():
                fn(*args, **kwargs)

            try:
                throttled.t.cancel()
            except(AttributeError):
                pass
            throttled.t = Timer(wait, call_it)
            throttled.t.start()

        return throttled

    return decorator


def as_dict(obj, recurse_list=None):
    # Primitive types
    if not hasattr(obj, "__dict__"):
        return str(obj)

    # Check if object was seen before, otherwise traverse it too
    if recurse_list is None:
        recurse_list = []
    if id(obj) in recurse_list:
        return "*loop detected*"
    else:
        recurse_list.append(id(obj))
        result = {}
        for key, val in obj.__dict__.items():

            if key.startswith("_"):
                continue

            if isinstance(val, list):
                if len(val) == 0:
                    element = str(val)
                else:
                    element = []
                    for item in val:
                        element.append(as_dict(item, recurse_list=recurse_list))
            else:
                element = as_dict(val, recurse_list=recurse_list)
            result[key] = element
        return result


# https://stackoverflow.com/a/47275100/3620725
def fix_pyqt():
    import sys

    sys._excepthook = sys.excepthook

    def my_exception_hook(exctype, value, traceback):
        print(exctype, value, traceback)
        sys._excepthook(exctype, value, traceback)
        sys.exit(1)

    sys.excepthook = my_exception_hook


# This makes it so PyQt5 windows don't become unresponsive in IPython outside app._exec() loops
def fix_ipython():
    from IPython import get_ipython
    ipython = get_ipython()
    if ipython is not None:
        ipython.magic("gui qt5")


def in_interactive_console():
    # https://stackoverflow.com/a/64523765/3620725
    return hasattr(sys, 'ps1')


def flatten_df(df):
    df = df.reset_index()
    df.columns = flatten_multiindex(df.columns)
    return df


def flatten_multiindex(mi, sep=" - ", format=None):
    import pandas as pd

    if issubclass(type(mi), pd.core.indexes.multi.MultiIndex):
        # Flatten multi-index headers
        if format == None:
            # Flatten by putting sep between each header value
            flat_index = [sep.join([str(x) for x in tup]).strip(sep)
                          for tup in mi.values]
        else:
            # Flatten according to the provided format string
            flat_index = []
            for tuple in mi.values:

                placeholders = []
                for name in mi.names:
                    if name is None:
                        name = ""
                    name = "{" + str(name) + "}"
                    placeholders.append(name)

                # Check if index segment contains each placeholder
                if all([item != "" for item in tuple]):
                    # Replace placeholders in format with corresponding values
                    flat_name = format
                    for i, val in enumerate(tuple):  # Iterates over the values in this index segment
                        flat_name = flat_name.replace(placeholders[i], val)
                else:
                    # If the segment doesn't contain all placeholders, just join them with sep instead
                    flat_name = sep.join(tuple).strip(sep)
                flat_index.append(flat_name)
    elif issubclass(type(mi), pd.core.indexes.base.Index):
        return mi
    else:
        raise TypeError(f"Expected Index but got {type(mi)}")

    return flat_index


# Alternative to dataframe.nunique that works when it contains unhashable items
def nunique(df):
    results = {}
    for col in df.columns:
        s = df[col]
        try:
            results[col] = s.nunique()
        except TypeError as e:
            results[col] = s.astype(str).nunique()

    return pd.Series(results)


# Alternative to series.unique that works when it contains unhashable items
def unique(s):
    try:
        return s.unique()
    except TypeError as e:
        return s.astype(str).unique()


def traverse_tree_widget(tree: Union[QtWidgets.QTreeWidget, QtWidgets.QTreeWidgetItem]) -> List[
    QtWidgets.QTreeWidgetItem]:
    if issubclass(type(tree), QtWidgets.QTreeWidget):
        tree = tree.invisibleRootItem()

    items = []
    for i in range(tree.childCount()):
        child = tree.child(i)
        sub_items = traverse_tree_widget(child)
        items += [child]
        items += sub_items

    return items


# Remove all widgets from a PyQt layout
def clear_layout(layout):
    if layout is not None:
        while layout.count():
            child = layout.takeAt(0)
            if child.widget() is not None:
                child.widget().deleteLater()
            elif child.layout() is not None:
                clear_layout(child.layout())


def unique_name(name, existing_names):
    if name in existing_names:
        for i in range(2, 999):
            new_name = f"{name} ({i})"
            if new_name not in existing_names:
                return new_name
        raise ValueError("Stopped generating unique name after 1000 attempts")
    else:
        return name


# Take a df and rename duplicate columns by appending number suffixes
def rename_duplicates(df):
    import copy
    new_columns = df.columns.values
    suffix = {key: 2 for key in set(new_columns)}
    dup = pd.Series(new_columns).duplicated()

    if type(df.columns) == pd.core.indexes.multi.MultiIndex:
        # Need to be mutable, make it list instead of tuples
        for i in range(len(new_columns)):
            new_columns[i] = list(new_columns[i])
        for ix, item in enumerate(new_columns):
            item_orig = copy.copy(item)
            if dup[ix]:
                for level in range(len(new_columns[ix])):
                    new_columns[ix][level] = new_columns[ix][level] + f"_{suffix[tuple(item_orig)]}"
                suffix[tuple(item_orig)] += 1

        # Convert back from list to tuples now that we're done mutating values
        for i in range(len(new_columns)):
            new_columns[i] = tuple(new_columns[i])

        df.columns = pd.MultiIndex.from_tuples(new_columns)
    # Not a MultiIndex
    else:
        for ix, item in enumerate(new_columns):
            if dup[ix]:
                new_columns[ix] = item + f"_{suffix[item]}"
                suffix[item] += 1
        df.columns = new_columns


def delete_datasets():
    from pandasgui.datasets import LOCAL_DATASET_DIR
    import shutil
    logger.info(f"Deleting sample dataset directory ({LOCAL_DATASET_DIR})")
    shutil.rmtree(LOCAL_DATASET_DIR)


def parse_date(s: pd.Series):
    try:
        return pd.to_datetime(s)
    except:
        return s


# Automatically try to parse dates for all columns
def parse_all_dates(df: Union[pd.DataFrame, pd.Series]):
    # Try to parse all string columns as dates
    if type(df) == pd.DataFrame:
        def parse_dates_if_str(col: pd.Series):
            if (col.dtype == object or col.dtype == pd.StringDtype):
                # Edge case where pd.to_datetime will work but we don't want it to
                if all(col.isna()):
                    return col
                else:
                    return parse_date(col)
            else:
                return col

        return df.apply(parse_dates_if_str)
    elif type(df) == pd.Series:
        return parse_date(df)


def clean_dataframe(df, name="DataFrame"):
    # Remove non-string column names
    converted_names = []
    if issubclass(type(df.columns), pd.core.indexes.multi.MultiIndex):
        levels = df.columns.levels
        for level in levels:
            if any([type(val) != str for val in level]):
                logger.warning(f"In {name}, converted MultiIndex level values to string in: {str(level)}")
                df.columns = df.columns.set_levels([[str(val) for val in level] for level in levels])
                converted_names.append(str(level))
        if converted_names:
            logger.warning(f"In {name}, converted MultiIndex level names to string: {', '.join(converted_names)}")
    else:
        for i, col in enumerate(df.columns):
            if type(col) != str:
                df.rename(columns={col: str(col)}, inplace=True)
                converted_names.append(str(col))
        if converted_names:
            logger.warning(f"In {name}, converted column names to string: {', '.join(converted_names)}")

    # Check for duplicate columns
    if any(df.columns.duplicated()):
        logger.warning(f"In {name}, renamed duplicate columns: {list(set(df.columns[df.columns.duplicated()]))}")
        rename_duplicates(df)

    return df


def test_logging():
    logger.debug("debug")
    logger.info("info")
    logger.warning("warning")
    logger.error("error")


# Resize a widget to a percentage of the screen size
def resize_widget(widget, x, y):
    from PyQt5 import QtCore, QtWidgets
    widget.resize(QtCore.QSize(int(x * QtWidgets.QDesktopWidget().screenGeometry().width()),
                               int(y * QtWidgets.QDesktopWidget().screenGeometry().height())))


def get_kwargs():
    frame = inspect.currentframe().f_back
    keys, _, _, values = inspect.getargvalues(frame)
    kwargs = {}
    for key in keys:
        if key != 'self':
            kwargs[key] = values[key]
    return kwargs


# Flatten nested iterables
def flatten_iter(item):
    t = []
    if type(item) in [list, tuple, set]:
        for sub_item in item:
            t += flatten_iter(sub_item)
    else:
        t.append(item)

    return t


# Make a string from a kwargs dict as they would be displayed when passed to a function
# eg. {'a': 5, 'b': 6} -> a=5, b=6
def kwargs_string(kwargs_dict):
    chunks = []
    for key, val in kwargs_dict.items():
        if isinstance(val, pd.DataFrame):
            chunk = f'{key}={val._pgdf.name}'
        else:
            chunk = f'{key}={repr(val)}'
        chunks.append(chunk)
    return ', '.join(chunks)


def get_function_body(func):
    import inspect
    from itertools import dropwhile
    source_lines = inspect.getsourcelines(func)[0]
    source_lines = dropwhile(lambda x: x.startswith('@'), source_lines)
    line = next(source_lines).strip()
    if not line.startswith('def '):
        return line.rsplit(':')[-1].strip()
    elif not line.endswith(':'):
        for line in source_lines:
            line = line.strip()
            if line.endswith(':'):
                break
    # Handle functions that are not one-liners
    first_line = next(source_lines)
    # Find the indentation of the first line
    indentation = len(first_line) - len(first_line.lstrip())
    return ''.join([first_line[indentation:]] + [line[indentation:] for line in source_lines])


# In North America, Week 1 of any given year is the week which contains January 1st. Weeks span Sunday to Saturday.
def get_week(timestamp):
    from datetime import datetime

    this_jan1 = datetime(timestamp.year, 1, 1)
    next_jan1 = datetime(timestamp.year + 1, 1, 1)
    day_of_year = int(timestamp.strftime('%j'))
    days_until_jan1 = (next_jan1 - timestamp).days

    # Sunday = 0
    def day_of_week(ts):
        return int(ts.strftime('%w'))

    # Check if timestamp is in the same week as next year's January
    if days_until_jan1 <= day_of_week(next_jan1):
        return 1
    else:
        # Week Number = ((Days since start of week 1) / 7) + 1
        return (day_of_year + day_of_week(this_jan1) - 1) // 7 + 1


def get_week_str(ts):
    year = ts.year
    week = get_week(ts)

    if ts.month == 12 and week == 1:
        year += 1

    return "{}W{:02}".format(year, week)


# Rename a variable in a Python expression
def refactor_variable(expr, old_name, new_name):
    import ast
    import astor

    tree = ast.parse(expr)
    for node in ast.walk(tree):
        if isinstance(node, ast.Name):
            if node.id == old_name:
                node.id = new_name

    return astor.code_gen.to_source(tree)


def get_figure_type(fig):
    # Plotly
    try:
        import plotly.basedatatypes
        if issubclass(type(fig), plotly.basedatatypes.BaseFigure):
            return "plotly"
    except ModuleNotFoundError:
        pass
    # Matplotlib
    try:
        import matplotlib.axes, matplotlib.figure
        if issubclass(type(fig), matplotlib.axes.Axes) \
                or issubclass(type(fig), matplotlib.figure.Figure):
            return "matplotlib"
    except ModuleNotFoundError:
        pass
    # Bokeh
    try:
        import bokeh.plotting
        if issubclass(type(fig), bokeh.plotting.Figure):
            return "bokeh"
    except ModuleNotFoundError:
        pass
    # Altair
    try:
        import altair.vegalite.v4.api
        if issubclass(type(fig), altair.vegalite.v4.api.Chart):
            return "altair"
    except ModuleNotFoundError:
        pass
    return None


# Take the text entered for a DataFrame cell and parse it into an appropriate type for the column
def parse_cell(text, column_dtype):
    import numpy as np
    if text == "":
        return np.nan

    if column_dtype == str:
        return text

    # Parse text using same logic as reading a CSV file by using a file buffer
    try:
        from io import StringIO
        import pandas as pd
        value = pd.read_csv(StringIO(text), dtype=column_dtype, header=None).values[0][0]
        return value
    except ValueError:
        raise ValueError(f"Could not convert {repr(text)} to type {column_dtype}")


value = parse_cell('nan', str)


def summarize_json(data, terse=True):
    from collections import defaultdict

    if len(data) == 0:
        print(f"Data is an empty {type(data).__name__}")
        return

    def list_keys(data, parent_key='ROOT'):
        parent_key = parent_key.strip('.')
        keys = []
        if isinstance(data, list):
            parent_subkeys = parent_key.split(".")
            for item in data:
                keys += list_keys(item, f'{".".join(parent_subkeys[:-1])}.[{parent_subkeys[-1]}]')
        elif isinstance(data, dict):
            for key, item in data.items():
                keys += list_keys(item, f'{parent_key}.{key}')
        else:
            return [f'{parent_key} - {type(data).__name__}']

        return keys

    keycount = defaultdict(lambda: 0)

    for key in reversed(sorted(list_keys(data))):
        keycount[key] += 1

    summary = ""

    width = max([len(str(count)) for count in keycount.values()])
    redundant_subkeys = set()
    for key, count in keycount.items():
        if terse:
            line = f"{count: <{width}} {key}"

            for redundant_subkey in sorted(list(redundant_subkeys), key=len, reverse=True):
                line = line.replace(redundant_subkey, ' ' * len(redundant_subkey))

            for i in range(line.count('.')):
                subkey = '.'.join(key.split('.')[:-1 - i])
                redundant_subkeys.add(subkey)
        else:
            line = f"{count: <{width}} {key}"

        summary += line + "\n"

    return summary


# Given two lists find pairs (src, dest) such that moving values in A according to each pair gives B
def get_movements(a: list, b: list):
    assert set(a) == set(b)

    from copy import copy
    a = copy(a)

    movements = []
    for dest, letter in enumerate(b):
        src = a.index(letter)
        if src != 0:
            movements.append((src + dest, dest))
        a.pop(src)
    return movements


event_lookup = {"0": "QEvent::None",
                "114": "QEvent::ActionAdded",
                "113": "QEvent::ActionChanged",
                "115": "QEvent::ActionRemoved",
                "99": "QEvent::ActivationChange",
                "121": "QEvent::ApplicationActivate",
                "122": "QEvent::ApplicationDeactivate",
                "36": "QEvent::ApplicationFontChange",
                "37": "QEvent::ApplicationLayoutDirectionChange",
                "38": "QEvent::ApplicationPaletteChange",
                "214": "QEvent::ApplicationStateChange",
                "35": "QEvent::ApplicationWindowIconChange",
                "68": "QEvent::ChildAdded",
                "69": "QEvent::ChildPolished",
                "71": "QEvent::ChildRemoved",
                "40": "QEvent::Clipboard",
                "19": "QEvent::Close",
                "200": "QEvent::CloseSoftwareInputPanel",
                "178": "QEvent::ContentsRectChange",
                "82": "QEvent::ContextMenu",
                "183": "QEvent::CursorChange",
                "52": "QEvent::DeferredDelete",
                "60": "QEvent::DragEnter",
                "62": "QEvent::DragLeave",
                "61": "QEvent::DragMove",
                "63": "QEvent::Drop",
                "170": "QEvent::DynamicPropertyChange",
                "98": "QEvent::EnabledChange",
                "10": "QEvent::Enter",
                "150": "QEvent::EnterEditFocus",
                "124": "QEvent::EnterWhatsThisMode",
                "206": "QEvent::Expose",
                "116": "QEvent::FileOpen",
                "8": "QEvent::FocusIn",
                "9": "QEvent::FocusOut",
                "23": "QEvent::FocusAboutToChange",
                "97": "QEvent::FontChange",
                "198": "QEvent::Gesture",
                "202": "QEvent::GestureOverride",
                "188": "QEvent::GrabKeyboard",
                "186": "QEvent::GrabMouse",
                "159": "QEvent::GraphicsSceneContextMenu",
                "164": "QEvent::GraphicsSceneDragEnter",
                "166": "QEvent::GraphicsSceneDragLeave",
                "165": "QEvent::GraphicsSceneDragMove",
                "167": "QEvent::GraphicsSceneDrop",
                "163": "QEvent::GraphicsSceneHelp",
                "160": "QEvent::GraphicsSceneHoverEnter",
                "162": "QEvent::GraphicsSceneHoverLeave",
                "161": "QEvent::GraphicsSceneHoverMove",
                "158": "QEvent::GraphicsSceneMouseDoubleClick",
                "155": "QEvent::GraphicsSceneMouseMove",
                "156": "QEvent::GraphicsSceneMousePress",
                "157": "QEvent::GraphicsSceneMouseRelease",
                "182": "QEvent::GraphicsSceneMove",
                "181": "QEvent::GraphicsSceneResize",
                "168": "QEvent::GraphicsSceneWheel",
                "18": "QEvent::Hide",
                "27": "QEvent::HideToParent",
                "127": "QEvent::HoverEnter",
                "128": "QEvent::HoverLeave",
                "129": "QEvent::HoverMove",
                "96": "QEvent::IconDrag",
                "101": "QEvent::IconTextChange",
                "83": "QEvent::InputMethod",
                "207": "QEvent::InputMethodQuery",
                "169": "QEvent::KeyboardLayoutChange",
                "6": "QEvent::KeyPress",
                "7": "QEvent::KeyRelease",
                "89": "QEvent::LanguageChange",
                "90": "QEvent::LayoutDirectionChange",
                "76": "QEvent::LayoutRequest",
                "11": "QEvent::Leave",
                "151": "QEvent::LeaveEditFocus",
                "125": "QEvent::LeaveWhatsThisMode",
                "88": "QEvent::LocaleChange",
                "176": "QEvent::NonClientAreaMouseButtonDblClick",
                "174": "QEvent::NonClientAreaMouseButtonPress",
                "175": "QEvent::NonClientAreaMouseButtonRelease",
                "173": "QEvent::NonClientAreaMouseMove",
                "177": "QEvent::MacSizeChange",
                "43": "QEvent::MetaCall",
                "102": "QEvent::ModifiedChange",
                "4": "QEvent::MouseButtonDblClick",
                "2": "QEvent::MouseButtonPress",
                "3": "QEvent::MouseButtonRelease",
                "5": "QEvent::MouseMove",
                "109": "QEvent::MouseTrackingChange",
                "13": "QEvent::Move",
                "197": "QEvent::NativeGesture",
                "208": "QEvent::OrientationChange",
                "12": "QEvent::Paint",
                "39": "QEvent::PaletteChange",
                "131": "QEvent::ParentAboutToChange",
                "21": "QEvent::ParentChange",
                "212": "QEvent::PlatformPanel",
                "217": "QEvent::PlatformSurface",
                "75": "QEvent::Polish",
                "74": "QEvent::PolishRequest",
                "123": "QEvent::QueryWhatsThis",
                "106": "QEvent::ReadOnlyChange",
                "199": "QEvent::RequestSoftwareInputPanel",
                "14": "QEvent::Resize",
                "204": "QEvent::ScrollPrepare",
                "205": "QEvent::Scroll",
                "117": "QEvent::Shortcut",
                "51": "QEvent::ShortcutOverride",
                "17": "QEvent::Show",
                "26": "QEvent::ShowToParent",
                "50": "QEvent::SockAct",
                "192": "QEvent::StateMachineSignal",
                "193": "QEvent::StateMachineWrapped",
                "112": "QEvent::StatusTip",
                "100": "QEvent::StyleChange",
                "87": "QEvent::TabletMove",
                "92": "QEvent::TabletPress",
                "93": "QEvent::TabletRelease",
                "171": "QEvent::TabletEnterProximity",
                "172": "QEvent::TabletLeaveProximity",
                "219": "QEvent::TabletTrackingChange",
                "22": "QEvent::ThreadChange",
                "1": "QEvent::Timer",
                "120": "QEvent::ToolBarChange",
                "110": "QEvent::ToolTip",
                "184": "QEvent::ToolTipChange",
                "194": "QEvent::TouchBegin",
                "209": "QEvent::TouchCancel",
                "196": "QEvent::TouchEnd",
                "195": "QEvent::TouchUpdate",
                "189": "QEvent::UngrabKeyboard",
                "187": "QEvent::UngrabMouse",
                "78": "QEvent::UpdateLater",
                "77": "QEvent::UpdateRequest",
                "111": "QEvent::WhatsThis",
                "118": "QEvent::WhatsThisClicked",
                "31": "QEvent::Wheel",
                "132": "QEvent::WinEventAct",
                "24": "QEvent::WindowActivate",
                "103": "QEvent::WindowBlocked",
                "25": "QEvent::WindowDeactivate",
                "34": "QEvent::WindowIconChange",
                "105": "QEvent::WindowStateChange",
                "33": "QEvent::WindowTitleChange",
                "104": "QEvent::WindowUnblocked",
                "203": "QEvent::WinIdChange",
                "126": "QEvent::ZOrderChange", }



import re
import time

import pkg_resources
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt


class FindToolbar(QtWidgets.QToolBar):
    def __init__(self, parent=None):
        """
        Creates modified toolbar with only a find textbox and match modification.

        Args:
            parent: parent widget (the gui) is required to access the current
                    DataFrame shown to the user.
        """
        super().__init__(parent=parent)

        # global variable initialization
        self.findThread = None
        # List of model QModelIndexes for indexes where a match was found
        self.search_matches = []
        # Current QModelIndex
        self.search_selection = None
        self.match_flags = {"regex": False, "case": False, "whole word": False}
        self.image_folder = "../resources/images"

        # main toolbar widget
        find_toolbar_widget = QtWidgets.QWidget()
        find_toolbar_layout = QtWidgets.QHBoxLayout()
        find_toolbar_layout.setContentsMargins(0, 0, 0, 0)
        find_toolbar_layout.setSpacing(0)
        find_toolbar_widget.setLayout(find_toolbar_layout)

        # textedit portion of toolbar
        self.find_textbox = ButtonLineEdit(content_margins=(0, 0, 5, 0))
        self.find_textbox.setPlaceholderText("Find")
        self.find_textbox.textChanged.connect(self.query)

        ####################
        # add match modification

        # add match case button
        match_case_icon_raw_path = self.image_folder + "/case-sensitive.svg"
        match_case_icon_path = pkg_resources.resource_filename(
            __name__, match_case_icon_raw_path
        )
        match_case_icon = QtGui.QIcon(match_case_icon_path)
        self.match_case_button = QtWidgets.QToolButton(self.find_textbox)
        self.match_case_button.setIcon(match_case_icon)
        self.match_case_button.setToolTip("Match Case")
        self.match_case_button.setCursor(Qt.PointingHandCursor)
        self.match_case_button.setCheckable(True)
        self.match_case_button.clicked.connect(self.toggle_match_case)
        self.find_textbox.add_button(self.match_case_button)

        # add match regex button
        regex_icon_raw_path = self.image_folder + "/regex.svg"
        regex_icon_path = pkg_resources.resource_filename(__name__, regex_icon_raw_path)
        regex_icon = QtGui.QIcon(regex_icon_path)
        self.match_regex_button = QtWidgets.QToolButton(self.find_textbox)
        self.match_regex_button.setIcon(regex_icon)
        self.match_regex_button.setToolTip("Use Regular Expression")
        self.match_regex_button.setCursor(Qt.PointingHandCursor)
        self.match_regex_button.setCheckable(True)
        self.match_regex_button.clicked.connect(self.toggle_regex)
        self.find_textbox.add_button(self.match_regex_button)

        # add match exactly button
        match_exactly_icon_raw_path = self.image_folder + "/match-exactly.svg"
        match_exactly_icon_path = pkg_resources.resource_filename(
            __name__, match_exactly_icon_raw_path
        )
        whole_word_icon = QtGui.QIcon(match_exactly_icon_path)
        self.match_exactly_button = QtWidgets.QToolButton(self.find_textbox)
        self.match_exactly_button.setIcon(whole_word_icon)
        self.match_exactly_button.setToolTip("Match Exactly")
        self.match_exactly_button.setCursor(Qt.PointingHandCursor)
        self.match_exactly_button.setCheckable(True)
        self.match_exactly_button.clicked.connect(self.toggle_match_exactly)
        self.find_textbox.add_button(self.match_exactly_button)
        find_toolbar_layout.addWidget(self.find_textbox)

        self.matches_found_label = QtWidgets.QLabel("Matches Found: 0")
        self.matches_found_label.setContentsMargins(0, 0, 7, 0)
        find_toolbar_layout.addWidget(self.matches_found_label)

        ####################
        # Main toolbar buttons

        # go to next match
        previous_match_button = QtWidgets.QPushButton()
        up_arrow_icon_raw_path = self.image_folder + "/arrow-up.svg"
        up_arrow_icon_path = pkg_resources.resource_filename(
            __name__, up_arrow_icon_raw_path)
        up_arrow_icon = QtGui.QIcon(up_arrow_icon_path)
        previous_match_button.setIcon(up_arrow_icon)
        previous_match_button.setToolTip("Previous match")
        previous_match_button.clicked.connect(self.select_previous_match)
        find_toolbar_layout.addWidget(previous_match_button)

        # go to previous match
        next_match_button = QtWidgets.QPushButton()
        down_arrow_icon_raw_path = self.image_folder + "/arrow-down.svg"
        down_arrow_icon_path = pkg_resources.resource_filename(
            __name__, down_arrow_icon_raw_path
        )
        down_arrow_icon = QtGui.QIcon(down_arrow_icon_path)
        next_match_button.setIcon(down_arrow_icon)
        next_match_button.setToolTip("Next match")
        next_match_button.clicked.connect(self.select_next_match)
        find_toolbar_layout.addWidget(next_match_button)

        # close find toolbar
        close_find_button = QtWidgets.QPushButton()
        cancel_icon_raw_path = self.image_folder + "/close.svg"
        cancel_icon_path = pkg_resources.resource_filename(
            __name__, cancel_icon_raw_path
        )
        close_icon = QtGui.QIcon(cancel_icon_path)
        close_find_button.setIcon(close_icon)
        close_find_button.setToolTip("Close Find Bar")
        close_find_button.clicked.connect(self.hide_find_bar)
        find_toolbar_layout.addWidget(close_find_button)

        self.addWidget(find_toolbar_widget)

        # hide toolbar
        self.setFixedHeight(0)

    @QtCore.pyqtSlot(str)
    def query(self, text):
        """
        Query operation done each time user changes the text.

        Args:
            text: text to search for.
        """
        # get current dataframe data
        current_pgdf = self.parent().store.selected_pgdf
        current_dataView = current_pgdf.dataframe_viewer.dataView
        current_model = current_dataView.model()
        df = current_pgdf.df

        # clear matches and selection from last search results
        if self.findThread:
            self.findThread.stop()
        self.search_matches = []
        current_dataView.selectionModel().clear()
        self.matches_found_label.setText("Matches Found: 0")
        self.search_selection = None

        if not text:
            return

        # Initialize findThread
        self.findThread = FindThread(df, text, self.match_flags)
        self.findThread.matches.connect(self.update_matches)
        self.findThread.start()

    @QtCore.pyqtSlot(list)
    def update_matches(self, cells_matched):
        """
        PyQt Slot that updates the matches found each time it gets a signal.

        Args:
            cells_matched: list of tuples - (row, col). Type QtCore.pyqtSignal(list).
        """

        current_pgdf = self.parent().store.selected_pgdf
        current_dataView = current_pgdf.dataframe_viewer.dataView
        current_model = current_dataView.model()

        # converts list of tuples to list of QtCore.QModelIndex for easy selection.
        match_idxs = [current_model.index(row, col) for row, col in cells_matched]
        self.search_matches.extend(match_idxs)

        matches_found_text = "Matches Found: " + str(len(self.search_matches))
        self.matches_found_label.setText(matches_found_text)

        if self.search_matches and self.search_selection is None:
            # highlight first match
            self.search_selection = 0
            self.highlight_match()

    @QtCore.pyqtSlot()
    def show_find_bar(self):
        if self.height() == 0:
            animation_duration = 200
            full_toolbar_height = 30

            showAnimation = QtCore.QVariantAnimation(self)
            showAnimation.setEasingCurve(QtCore.QEasingCurve.InOutQuad)
            showAnimation.setDuration(animation_duration)
            showAnimation.setStartValue(0)
            showAnimation.setEndValue(full_toolbar_height)
            showAnimation.valueChanged.connect(lambda val: self.setFixedHeight(val))

            showAnimation.start()

            # clear the last search, and set cursor on the QLineEdit
            self.find_textbox.setText("")
            self.find_textbox.setFocus()

    @QtCore.pyqtSlot()
    def hide_find_bar(self):
        if self.height() == 30:
            animation_duration = 200
            full_toolbar_height = 30

            hideAnimation = QtCore.QVariantAnimation(self)
            hideAnimation.setEasingCurve(QtCore.QEasingCurve.InOutQuad)
            hideAnimation.setDuration(animation_duration)
            hideAnimation.setStartValue(full_toolbar_height)
            hideAnimation.setEndValue(0)
            hideAnimation.valueChanged.connect(lambda val: self.setFixedHeight(val))

            hideAnimation.start()

            # stop any running findThread
            if self.findThread:
                self.findThread.stop()

    @QtCore.pyqtSlot()
    def select_next_match(self):
        if self.search_matches:
            # loop around to the first match if user hits last match
            if self.search_selection == len(self.search_matches) - 1:
                self.search_selection = 0
            else:
                self.search_selection += 1

            self.highlight_match()

    @QtCore.pyqtSlot()
    def select_previous_match(self):
        if self.search_matches:
            # loop around to the last match if user hits first match
            if self.search_selection == 0:
                self.search_selection = len(self.search_matches) - 1
            else:
                self.search_selection -= 1

            self.highlight_match()

    @QtCore.pyqtSlot()
    def toggle_match_case(self):
        self.match_flags["case"] = not self.match_flags["case"]

        if self.find_textbox.text():
            self.query(self.find_textbox.text())

    @QtCore.pyqtSlot()
    def toggle_regex(self):
        self.match_flags["regex"] = not self.match_flags["regex"]
        if self.match_flags["whole word"]:
            self.match_flags["whole word"] = False
            self.match_exactly_button.setChecked(False)

        if self.find_textbox.text():
            self.query(self.find_textbox.text())

    @QtCore.pyqtSlot()
    def toggle_match_exactly(self):
        self.match_flags["whole word"] = not self.match_flags["whole word"]
        if self.match_flags["regex"]:
            self.match_flags["regex"] = False
            self.match_regex_button.setChecked(False)

        if self.find_textbox.text():
            self.query(self.find_textbox.text())

    def highlight_match(self):
        # clear last seletion
        current_pgdf = self.parent().store.selected_pgdf
        current_dataView = current_pgdf.dataframe_viewer.dataView
        current_dataView.selectionModel().select(
            self.search_matches[self.search_selection],
            QtCore.QItemSelectionModel.ClearAndSelect,
        )
        current_dataView.scrollTo(self.search_matches[self.search_selection])


class ButtonLineEdit(QtWidgets.QLineEdit):
    buttonClicked = QtCore.pyqtSignal(bool)

    def __init__(self, parent=None, content_margins=(0, 0, 0, 0)):
        """
        Modified LineEdit with buttons inside the LineEdit.
        """
        super(ButtonLineEdit, self).__init__(parent)

        self.setContentsMargins(*content_margins)
        self.buttons = []

    def resizeEvent(self, event):
        """
        Moves all buttons to the correct position or else the buttons will
        be stacked on top of each other on the left.
        """
        frameWidth = self.style().pixelMetric(QtWidgets.QStyle.PM_DefaultFrameWidth)
        right_margin = self.getContentsMargins()[2]
        bottom_margin = self.getContentsMargins()[3]

        for i, button in enumerate(self.buttons):
            buttonSize = button.sizeHint()
            button.move(
                self.rect().right()
                - frameWidth
                - buttonSize.width() * (i + 1)
                - right_margin,
                (self.rect().bottom() - buttonSize.height() + 1) / 2 - bottom_margin,
            )
        super(ButtonLineEdit, self).resizeEvent(event)

    def add_button(self, button):
        """
        Adds a button to the right side of the QLineEdit.

        Args:
            button: Type QtWidgets.QPushButton or QtWidgets.QToolButton
        """
        self.buttons.append(button)

        # makes sure text doesn't type behind the buttons
        totalWidth = sum([b.sizeHint().width() for b in self.buttons])
        frameWidth = self.style().pixelMetric(QtWidgets.QStyle.PM_DefaultFrameWidth)

        # makes sure the typing area doesn't get too small if toolbar is shrunk.
        maxHeight = max([b.sizeHint().height() for b in self.buttons])
        self.setMinimumSize(
            max(self.minimumSizeHint().width(), totalWidth + frameWidth * 2 + 2),
            max(self.minimumSizeHint().height(), maxHeight + frameWidth * 2 + 2),
        )


class FindThread(QtCore.QThread):
    matches = QtCore.pyqtSignal(list)

    def __init__(self, df, text, match_flags, parent=None):
        """
        Thread to search DataFrame for a string.

        Args:
            df: Type pd.DataFrame
            text: Text to search for. Type string.
            match_flags: User enabled match flags. Can match case, regex, or exact.
                         Type dict.
        """
        QtCore.QThread.__init__(self, parent=parent)
        self.isRunning = True
        self.df = df
        self.text = text
        self.match_flags = match_flags
        self.max_chunk_size = 10000
        self.chunks = self.split_chunks()

    def split_chunks(self):
        """
        Splits each Series in the DataFrame into many Series.
        Number of chunks outputted depends on size of each column
        and self.max_chunk_size.

        Returns:
            chunks: List of pd.Series.
        """
        chunks = []
        for col_idx, col_name in enumerate(self.df.columns):
            column = self.df[col_name].copy()
            while len(column) > 0:
                chunk = column.iloc[: self.max_chunk_size]
                chunks.append(chunk)
                column = column.iloc[self.max_chunk_size:]
        return chunks

    def get_matches(self, chunk):
        """
        Gets row numbers of matches.

        Args:
            chunk: Type pd.Series
        """
        if self.match_flags["whole word"]:
            if self.match_flags["case"]:
                rows_with_match = chunk[chunk.astype(str) == self.text]
            else:
                rows_with_match = chunk[
                    chunk.astype(str).str.lower() == self.text.lower()
                    ]
        else:
            pd_match_flags = self.match_flags.copy()
            pd_match_flags.pop("whole word")
            check_for_match = chunk.astype(str).str.contains(
                self.text, **pd_match_flags
            )
            rows_with_match = chunk[check_for_match]

        return rows_with_match

    def run(self):
        col_idx = 0
        # since the chunks don't preserve row number, need to keep track of
        # chunks processed to get row number
        chunks_in_column = 0
        for chunk in self.chunks:
            try:
                # gives series with dtype bool
                rows_with_match = self.get_matches(chunk)
                # gives indicies where the series is True
                row_idx_with_match = [
                    chunk.index.get_loc(i) + chunks_in_column * self.max_chunk_size
                    for i in rows_with_match.index
                ]
                # output list of table coordinates where match is found
                coords_with_match = [
                    (row_idx, col_idx) for row_idx in row_idx_with_match
                ]

                # check if a stop is requested
                # (in case the user types another letter)
                if self.isRunning:
                    self.matches.emit(coords_with_match)
                else:
                    break

                chunks_in_column += 1

                # if the last index of the chunk is equal to the last index
                # of the whole dataframe, move to the next column
                if chunk.index[-1] == self.df.index[-1]:
                    col_idx += 1
                    chunks_in_column = 0

                # waits 50 milliseconds to process gui interactions,
                # so the gui does not freeze
                time.sleep(0.05)
            except re.error:
                break
        self.isRunning = False

    def stop(self):
        self.isRunning = False
        self.quit()
        self.wait()



from math import ceil

from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *


class Spinner(QWidget):
    mColor = QColor(Qt.gray)
    mRoundness = 100.0
    mMinimumTrailOpacity = 31.4159265358979323846
    mTrailFadePercentage = 50.0
    mRevolutionsPerSecond = 1.57079632679489661923
    mNumberOfLines = 20
    mLineLength = 10
    mLineWidth = 2
    mInnerRadius = 20
    mCurrentCounter = 0
    mIsSpinning = False

    def __init__(
        self, centerOnParent=True, disableParentWhenSpinning=True, *args, **kwargs
    ):
        QWidget.__init__(self, *args, **kwargs)
        self.mCenterOnParent = centerOnParent
        self.mDisableParentWhenSpinning = disableParentWhenSpinning
        self.initialize()

    def initialize(self):
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.rotate)
        self.updateSize()
        self.updateTimer()
        self.hide()

    @pyqtSlot()
    def rotate(self):
        self.mCurrentCounter += 1
        if self.mCurrentCounter > self.numberOfLines():
            self.mCurrentCounter = 0
        self.update()

    def updateSize(self):
        size = (self.mInnerRadius + self.mLineLength) * 2
        self.setFixedSize(size, size)

    def updateTimer(self):
        self.timer.setInterval(
            1000 / (self.mNumberOfLines * self.mRevolutionsPerSecond)
        )

    def updatePosition(self):
        if self.parentWidget() and self.mCenterOnParent:
            self.move(
                self.parentWidget().width() / 2 - self.width() / 2,
                self.parentWidget().height() / 2 - self.height() / 2,
            )

    def lineCountDistanceFromPrimary(self, current, primary, totalNrOfLines):
        distance = primary - current
        if distance < 0:
            distance += totalNrOfLines
        return distance

    def currentLineColor(
        self, countDistance, totalNrOfLines, trailFadePerc, minOpacity, color
    ):
        if countDistance == 0:
            return color

        minAlphaF = minOpacity / 100.0

        distanceThreshold = ceil((totalNrOfLines - 1) * trailFadePerc / 100.0)
        if countDistance > distanceThreshold:
            color.setAlphaF(minAlphaF)

        else:
            alphaDiff = self.mColor.alphaF() - minAlphaF
            gradient = alphaDiff / distanceThreshold + 1.0
            resultAlpha = color.alphaF() - gradient * countDistance
            resultAlpha = min(1.0, max(0.0, resultAlpha))
            color.setAlphaF(resultAlpha)
        return color

    def paintEvent(self, event):
        self.updatePosition()
        painter = QPainter(self)
        painter.fillRect(self.rect(), Qt.transparent)
        painter.setRenderHint(QPainter.Antialiasing, True)
        if self.mCurrentCounter > self.mNumberOfLines:
            self.mCurrentCounter = 0
        painter.setPen(Qt.NoPen)

        for i in range(self.mNumberOfLines):
            painter.save()
            painter.translate(
                self.mInnerRadius + self.mLineLength,
                self.mInnerRadius + self.mLineLength,
            )
            rotateAngle = 360.0 * i / self.mNumberOfLines
            painter.rotate(rotateAngle)
            painter.translate(self.mInnerRadius, 0)
            distance = self.lineCountDistanceFromPrimary(
                i, self.mCurrentCounter, self.mNumberOfLines
            )
            color = self.currentLineColor(
                distance,
                self.mNumberOfLines,
                self.mTrailFadePercentage,
                self.mMinimumTrailOpacity,
                self.mColor,
            )
            painter.setBrush(color)
            painter.drawRoundedRect(
                QRect(0, -self.mLineWidth // 2, self.mLineLength, self.mLineLength),
                self.mRoundness,
                Qt.RelativeSize,
            )
            painter.restore()

    def start(self):
        self.updatePosition()
        self.mIsSpinning = True

        if self.parentWidget() and self.mDisableParentWhenSpinning:
            self.parentWidget().setEnabled(False)

        if not self.timer.isActive():
            self.timer.start()
            self.mCurrentCounter = 0

    def stop(self):
        self.mIsSpinning = False
        self.hide()

        if self.parentWidget() and self.mDisableParentWhenSpinning:
            self.parentWidget().setEnabled(True)

        if self.timer.isActive():
            self.timer.stop()
            self.mCurrentCounter = 0

    def setNumberOfLines(self, lines):
        self.mNumberOfLines = lines
        self.updateTimer()

    def setLineLength(self, length):
        self.mLineLength = length
        self.updateSize()

    def setLineWidth(self, width):
        self.mLineWidth = width
        self.updateSize()

    def setInnerRadius(self, radius):
        self.mInnerRadius = radius
        self.updateSize()

    def color(self):
        return self.mColor

    def roundness(self):
        return self.mRoundness

    def minimumTrailOpacity(self):
        return self.mMinimumTrailOpacity

    def trailFadePercentage(self):
        return self.mTrailFadePercentage

    def revolutionsPersSecond(self):
        return self.mRevolutionsPerSecond

    def numberOfLines(self):
        return self.mNumberOfLines

    def lineLength(self):
        return self.mLineLength

    def lineWidth(self):
        return self.mLineWidth

    def innerRadius(self):
        return self.mInnerRadius

    def isSpinning(self):
        return self.mIsSpinning

    def setRoundness(self, roundness):
        self.mRoundness = min(0.0, max(100, roundness))

    def setColor(self, color):
        self.mColor = color

    def setRevolutionsPerSecond(self, revolutionsPerSecond):
        self.mRevolutionsPerSecond = revolutionsPerSecond
        self.updateTimer()

    def setTrailFadePercentage(self, trail):
        self.mTrailFadePercentage = trail

    def setMinimumTrailOpacity(self, minimumTrailOpacity):
        self.mMinimumTrailOpacity = minimumTrailOpacity


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    dial = QDialog()
    w = Spinner(dial)
    dial.show()
    w.start()
    QTimer.singleShot(1000, w.stop)
    sys.exit(app.exec_())



import os
import sys
import tempfile

from PyQt5 import QtCore, QtGui, QtWidgets, sip
from PyQt5.QtCore import Qt
import PyQt5
import logging

from pandasgui.store import PandasGuiStoreItem
from pandasgui.utility import get_figure_type

logger = logging.getLogger(__name__)

# https://stackoverflow.com/a/64743807/3620725
os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] = "--enable-logging --log-level=3"

# Since pandasgui might be imported after other packages already created a QApplication,
# we need to hack around this import restriction on QtWebEngineWidgets
# https://stackoverflow.com/a/57436077/3620725

if "PyQt5.QtWebEngineWidgets" not in sys.modules:
    app = QtWidgets.QApplication.instance()

    if app is None:
        from PyQt5 import QtWebEngineWidgets
    else:
        logger.warning("Reinitializing existing QApplication to allow import of QtWebEngineWidgets. "
                       "This may cause problems. "
                       "To avoid this, import pandasgui or PyQt5.QtWebEngineWidgets before a QApplication is created.")
        app.quit()
        sip.delete(app)
        from PyQt5 import QtWebEngineWidgets

        app.__init__(sys.argv + ["--ignore-gpu-blacklist", "--enable-gpu-rasterization"])


class FigureViewer(PyQt5.QtWebEngineWidgets.QWebEngineView, PandasGuiStoreItem):
    def __init__(self, fig=None, store=None):
        super().__init__()
        self.store = store
        self.page().profile().downloadRequested.connect(self.on_downloadRequested)

        # Fix scrollbar sometimes disappearing after Plotly autosizes and leaving behind white space
        self.settings().setAttribute(self.settings().ShowScrollBars, False)
        self.settings().setAttribute(PyQt5.QtWebEngineWidgets.QWebEngineSettings.WebGLEnabled, True)

        # https://stackoverflow.com/a/8577226/3620725
        self.temp_file = tempfile.NamedTemporaryFile(mode="w", suffix=".html", delete=False)
        self.set_figure(fig)

        self.setWindowTitle("Plotly Viewer")

    def set_figure(self, fig=None):

        self.temp_file.seek(0)
        dark = self.store is not None and self.store.settings.theme.value == "dark"
        fig_type = get_figure_type(fig)

        if fig is None:
            html = ""
        elif fig_type == "plotly":
            from plotly.io import to_html
            if dark:
                fig.update_layout(template="plotly_dark", autosize=True)
            html = to_html(fig, config={"responsive": True})
        elif fig_type == "bokeh":
            from bokeh.resources import CDN
            from bokeh.embed import file_html
            html = file_html(fig, resources=CDN, title="my fig")
        elif fig_type == "matplotlib":
            fig = fig.get_figure() or fig
            import base64
            from io import BytesIO
            tmpfile = BytesIO()
            fig.savefig(tmpfile, format='png')
            encoded = base64.b64encode(tmpfile.getvalue()).decode('utf-8')
            html = '<img src=\'data:image/png;base64,{}\'>'.format(encoded)
        elif fig_type == "altair":
            fig = fig.properties(
                width='container',
                height='container'
            )
            from io import StringIO
            tmp = StringIO()
            fig.save(tmp, format='html')
            html = tmp.getvalue()

        else:
            raise TypeError

        html = html.replace("<style>",
                     "<style>"
                     "body{margin: 0; width:100vw; height:100vh;} "
                     # https://github.com/vega/react-vega/issues/85#issuecomment-795138175
                     ".vega-embed{width:100%; height:100%;} "
                     )

        self.temp_file.write(html)
        self.temp_file.truncate()
        self.temp_file.seek(0)
        self.load(QtCore.QUrl.fromLocalFile(self.temp_file.name))

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        self.temp_file.close()
        os.unlink(self.temp_file.name)
        super().closeEvent(event)

    # https://stackoverflow.com/questions/55963931/how-to-download-csv-file-with-qwebengineview-and-qurl
    def on_downloadRequested(self, download):
        dialog = QtWidgets.QFileDialog()
        dialog.setDefaultSuffix(".png")
        path, _ = dialog.getSaveFileName(self, "Save File", os.path.join(os.getcwd(), "newplot.png"), "*.png")
        if path:
            download.setPath(path)
            download.accept()

    def pg_widget(self):
        return self

if __name__ == "__main__":
    # Create a QtWidgets.QApplication instance or use the existing one if it exists
    app = QtWidgets.QApplication(sys.argv)

    import numpy as np
    import plotly.graph_objs as go
    from pandasgui.utility import fix_ipython, fix_pyqt

    fix_ipython()
    fix_pyqt()

    fig = go.Figure()
    fig.add_scatter(
        x=np.random.rand(100),
        y=np.random.rand(100),
        mode="markers",
        marker={
            "size": 30,
            "color": np.random.rand(100),
            "opacity": 0.6,
            "colorscale": "Viridis",
        },
    )

    pv = FigureViewer(fig)
    pv.show()
    app.exec_()



def enable_profiling(glob_call_time_stack=[]):
    import sys
    import inspect
    import os
    import time

    def tracefunc(frame, event, arg,
                  indent=[0], filename=__file__, call_time_stack=glob_call_time_stack, last_line_num=[None],
                  last_line_start=[None]):
        '''
        frame: See frame in table https://docs.python.org/3/library/inspect.html#types-and-members
        event: https://docs.python.org/3/library/inspect.html#types-and-members
        arg: For 'return' event, arg is return value.
             For 'exception' event, arg is (exception, value, traceback).
             Otherwise arg is None
        '''
        global TRACEFUNC_PROJECT_FILES
        if 'TRACEFUNC_PROJECT_FILES' not in globals():
            caller_file = inspect.currentframe().f_code.co_filename
            dirname = os.path.dirname(caller_file)
            # Find project root path
            while True:
                if '__init__.py' in os.listdir(dirname):
                    project_root = dirname
                    break
                elif os.path.dirname(dirname) == dirname:
                    # Got to drive root without __init__ file so project root is just the folder containing the caller file
                    project_root = os.path.dirname(caller_file)
                    break
                dirname = os.path.dirname(dirname)

            TRACEFUNC_PROJECT_FILES = []
            for root, subFolder, files in os.walk(project_root):
                for item in files:
                    if item.endswith(".py"):
                        path = str(os.path.join(root, item)).replace('\\', '/')
                        TRACEFUNC_PROJECT_FILES.append(path)

        line = frame.f_lineno
        file_path = frame.f_code.co_filename
        file_name = os.path.basename(file_path)
        object_name = frame.f_code.co_name

        if file_path in TRACEFUNC_PROJECT_FILES:
            if event == "call":
                call_time_stack.append(time.time())
                indent[0] += 3
                print("-" * indent[0] + f"> call function {frame.f_code.co_name} [Line {frame.f_lineno}]")
            elif event == "return":
                print("<" + "-" * indent[0], "exit function", frame.f_code.co_name,
                      f"({time.time() - call_time_stack.pop():.2f}s)")
                indent[0] -= 3
            if event == 'line':
                this_line_start = time.time()
                this_line_num = frame.f_lineno
        return tracefunc

    sys.settrace(tracefunc)
